"""
Monkey-patch for numpy to use czipfile rather than zipfile
"""
from __future__ import absolute_import

import imp
import sys

import zipfile


def inject_czipfile():
    _inject('zipfile','czipfile')


def inject_zipfile():
    _inject('zipfile','zipfile')

def _inject(module, new_module):
    if imp.lock_held() is True:
        if sys.modules.get(module):
            del sys.modules[module]
        sys.modules[new_module] = __import__(new_module)
        sys.modules[module] = __import__(new_module)

def use_czipfile(f):
    def wrapper(*args, **kwargs):
        inject_czipfile()
        value = f(*args, **kwargs)
        inject_zipfile()
        return value

    return wrapper




#!/usr/bin/env python

import nltk
import re
import os
from  unidecode import unidecode
from translate import Translator as Ts
from vsm.corpus.util import *

"""
Uses `translate` python module from https://pypi.python.org/pypi/translate
"""

def sent_tokenize(text, lang='english'):
    tokenizer = nltk.data.load('tokenizers/punkt/{0}.pickle'.format(lang))
    return tokenizer.tokenize(text)

def cleanup(s):
    s = rehyph(s)
    s = s.strip('\n')
    def replace(match):
        if match:
            if match.group(0).startswith(r'\*'):
                return unidecode(match.group(0))
        else:
            return ''

    return re.sub(r"[\x90-\xff]", replace, s)


def transwrapper(text, from_lang, to_lang):
    
    if from_lang == 'en':
        lang = 'english'
    elif from_lang == 'fr':
        lang = 'french'
    elif from_lang == 'de':
        lang = 'german'
    sli = sent_tokenize(text, lang=lang)
    
    out = ''
    for sent in sli:
        sent = cleanup(sent) 
        
        ts = Ts(from_lang=from_lang, to_lang=to_lang)
        target = ts.translate(sent)
        out += target
    
    return out


if __name__=="__main__":
    frompath = 'darwin-de/'
    topath = 'darwin-de-translate/'
    
    books = os.listdir(frompath)
    books.sort()
    
    for book in books:
        book_path = os.path.join(frompath, book)
        print book_path
        pages = os.listdir(book_path)
        pages.sort()
        
        for page in pages:
            page_name = os.path.join(book_path, page)
            
            with open(page_name, 'r') as f:
                try:
                    out = transwrapper(f.read(), 'de', 'en')
                    out = out.encode('utf-8')
                except:
                    out = ''
                    print page_name, ' failed translation.'
                
                try:
                    os.mkdir(os.path.join(topath, book))
                except OSError:
                    pass
                topage = os.path.join(topath, book, page)
                with open(topage, 'w') as fout:
    """ 
    # for individual pages 
    fin = 'darwin-de/wu.89101307601/00000636.txt'
    fout = 'darwin-de-translate/wu.89101307601/00000636.txt'

    with open(fin, 'r') as f:
        out = transwrapper(f.read(), 'de', 'en')
        out = out.encode('utf-8')
        with open(fout, 'w') as fo:
            fo.write(out)
                fout.write(out)"""



import numpy as np

from vsm.model.base import BaseModel


__all__ = ['BeagleEnvironment']


class BeagleEnvironment(BaseModel):
    """
    `BeagleEnvironment` is a randomly generated fixed vectors
    representing the environment.
    """
    
    def __init__(self, corpus, n_cols=2048, dtype=np.float64, 
                 context_type='sentence'):
        """
        Initialize BeagleEnvironment.

        :param corpus: Source of observed data.
        :type corpus: Corpus

        :param n_cols: Number of columns. Default is 2048.
        :type n_cols: int, optional

        :param dtype: Numpy dtype for matrix attribute. Default is `np.float64`.
        :type dtype: np.dtype, optional

        :param context_type: Name of tokenization stored in `corpus` whose
            tokens will be treated as documents. Default is `sentence`.
        :type context_type: string, optional
        """
        self.context_type = context_type
        self.shape = (corpus.words.shape[0], n_cols)
        self.dtype = dtype


    def train(self):
        """
        Sets a m x n environment matrix where m is the number of words in
        `corpus` and n is `n_cols`. The matrix consists of randomly generated
        vectors. 
        """
        self.matrix = np.array(np.random.normal(size=self.shape),
                               dtype=self.dtype)
        # normalize rows
        self.matrix /= np.sqrt((self.matrix * self.matrix).sum(1)[:,np.newaxis])





# cython: binding=True
# cython: wraparound=False
# cython: boundscheck=False
# cython: cdivision=True

import cython
cimport cython

import numpy as np
cimport numpy as np

ctypedef np.float32_t NP_FLOAT_t
#TODO: figure out how to use np types in python code
#ctypedef fused NP_FLOAT_t:
#    np.float32_t
#    np.float64_t

ctypedef fused CORPUS_t:
    unsigned int
    unsigned short

ctypedef fused TOPIC_t:
    unsigned short
    unsigned char

cdef extern from "math.h":
    float logf(float n)

@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
def cgs_update(int itr, 
               CORPUS_t [:] corpus,
               np.ndarray[NP_FLOAT_t, ndim=2] word_top,
               np.ndarray[NP_FLOAT_t] inv_top_sums,
               np.ndarray[NP_FLOAT_t, ndim=2] top_doc,
               TOPIC_t [:] Z,
               int [:] indices,
               str mtrand_str,
               unsigned int [:] mtrand_keys,
               int mtrand_pos,
               int mtrand_has_gauss,
               float mtrand_cached_gaussian):
    
    cdef int first, last
    cdef long stop, doc_len, offset
    cdef NP_FLOAT_t r, s
    cdef Py_ssize_t i, j, idx, w, t, k

    cdef int V = corpus.shape[0]
    cdef int N = indices.shape[0]
    cdef int K = word_top.shape[1]
    cdef int W = word_top.shape[0]
    
    cdef NP_FLOAT_t log_p = 0
    cdef np.ndarray[NP_FLOAT_t, ndim=2] log_wk = np.log(word_top * inv_top_sums)
    cdef np.ndarray[NP_FLOAT_t, ndim=2] log_kd = np.log(top_doc / top_doc.sum(0))

    cdef object np_random_state = np.random.RandomState()
    np_random_state.set_state((mtrand_str, mtrand_keys, 
                               mtrand_pos, mtrand_has_gauss, 
                               mtrand_cached_gaussian))
    cdef np.ndarray[NP_FLOAT_t] samples = np_random_state.uniform(size=V).astype(np.float32)
    cdef np.ndarray[NP_FLOAT_t] dist = np.zeros((K,), dtype=np.float32)

    cdef object mtrand_state = np_random_state.get_state()


    with nogil:
        for i in range(N):
    
            if i==0:
                doc_len = indices[0]
                offset = 0
            else:
                offset = indices[i-1]
                stop = indices[i]
                doc_len = stop - offset 
    
            for j in range(doc_len):
    
                idx = offset + j
                w,k = corpus[idx], Z[idx]
    
                log_p += log_wk[w, k] + log_kd[k, i]
    
                if itr > 0:
                    word_top[w, k] -= 1
                    s = inv_top_sums[k]
                    inv_top_sums[k] = s / (1 - s)
                    top_doc[k, i] -= 1
    
                t = 0
                dist[t] = <NP_FLOAT_t>(inv_top_sums[t] * word_top[w,t] * top_doc[t,i])
                for t in range(1,K):
                    dist[t] = dist[t-1] + <NP_FLOAT_t>(inv_top_sums[t] * word_top[w,t] * top_doc[t,i])
                
                r = samples[idx] * dist[K-1]
                for k in range(K):
                    if r < dist[k]:
                        break
                """
                # This code implements binary search for the right insertion
                # point for the probability in the cumulative distribution
                first = 0
                last = K - 1
                while first < last:
                    k = (first + last) / 2
                    if r < dist[k]:
                        last = k
                    else:
                        first = k + 1
                """
    
                word_top[w, k] += 1
                s = inv_top_sums[k]
                inv_top_sums[k] = s / (1 + s) 
                top_doc[k, i] += 1
    
                Z[idx] = <TOPIC_t>(k)
            
    return (np.asarray(word_top), np.asarray(inv_top_sums), 
            np.asarray(top_doc), np.asarray(Z), log_p, 
            mtrand_state[0], mtrand_state[1], mtrand_state[2], 
            mtrand_state[3], mtrand_state[4])



from __future__ import division
from __future__ import absolute_import
from builtins import zip
from builtins import object
from past.utils import old_div
import numpy as np

from vsm.spatial import angle
from vsm.structarr import *
from vsm.exceptions import *
from .types import *
from .labeleddata import *
from .wrappers import *


__all__ = ['LsaViewer']


class LsaViewer(object):
    """
    A class for viewing LSA model.
    """
    
    def __init__(self, corpus, model):
        """
        Initialize LsaViewer.

        :param corpus: Source of observed data.
        :type corpus: :class:`Corpus`

        :param model: An LSA model.
        :type model: Lsa
        """
        self.corpus = corpus
        self.model = model

    @deprecated_meth("dist_word_word")
    def sim_word_word(self, word_or_words, weights=[], 
                       filter_nan=True, print_len=10, as_strings=True, 
                       dist_fn=angle, order='i'):

        pass


    def dist_word_word(self, word_or_words, weights=[], 
                       filter_nan=True, print_len=10, as_strings=True, 
                       dist_fn=angle, order='i'):
        """
        Computes and sorts the distances between a word or list
        of words and every word.

        :param word_or_words: Query word(s) to which distances are calculated.
        :type word_or_words: string or list of strings
        
        :param weights: Specify weights for each query word in `word_or_words`. 
            Default uses equal weights (i.e. arithmetic mean)
        :type weights: list of floating point, optional
        
        :param filter_nan: If `True` not a number entries are filtered.
            Default is `True`.
        :type filter_nan: boolean, optional

        :param print_len: Number of words to be displayed. Default is 10.
        :type print_len: int, optional

        :param as_strings: If `True`, returns a list of words as strings rather
            than their integer representations. Default is `True`.
        :type as_strings: boolean, optional
        
        :param dist_fn: A distance function from functions in vsm.spatial. 
            Default is :meth:`angle`.
        :type dist_fn: string, optional
        
        :param order: Order of sorting. 'i' for increasing and 'd' for
            decreasing order. Default is 'i'.
        :type order: string, optional

        :returns: an instance of :class:`LabeledColumn`.
            A 2-dim array containing words and their distances to 
            `word_or_words`. 
        
        :See Also: :meth:`vsm.viewer.wrappers.dist_word_word`
        """
        return dist_word_word(word_or_words, self.corpus, 
                                self.model.word_matrix.T, weights=weights, 
                                filter_nan=filter_nan, 
                                print_len=print_len, as_strings=True, 
                                dist_fn=dist_fn, order=order)

    @deprecated_meth("dist_doc_doc")
    def sim_doc_doc(self, doc_or_docs, weights=[], print_len=10, 
                     filter_nan=True, label_fn=def_label_fn, as_strings=True,
                     dist_fn=angle, order='i'):
        pass
        
    def dist_doc_doc(self, doc_or_docs, weights=[], print_len=10, 
                     filter_nan=True, label_fn=def_label_fn, as_strings=True,
                     dist_fn=angle, order='i'):
        """
        Computes and sorts the distances between a
        document or list of documents and every document.

        :param doc_or_docs: Query document(s) to which distances
            are calculated.
        :type doc_or_docs: string/integer or list of strings/integers
        
        :param weights: Specify weights for each query doc in `doc_or_docs`. 
            Default uses equal weights (i.e. arithmetic mean)
        :type weights: list of floating point, optional
        
        :param print_len: Number of words to be displayed. Default is 10.
        :type print_len: int, optional

        :param filter_nan: If `True` not a number entries are filtered.
            Default is `True`.
        :type filter_nan: boolean, optional
 
        :param label_fn: A function that defines how documents are represented.
            Default is def_label_fn which retrieves the labels from corpus
            metadata.
        :type label_fn: string, optional
        
        :param as_strings: If `True`, returns a list of words rather than
            their integer representations. Default is `True`.
        :type as_strings: boolean, optional
        
        :param dist_fn: A distance function from functions in vsm.spatial. 
            Default is :meth:`angle`.
        :type dist_fn: string, optional
        
        :param order: Order of sorting. 'i' for increasing and 'd' for
            decreasing order. Default is 'i'.
        :type order: string, optional

        :returns: an instance of :class:`LabeledColumn`.
            A 2-dim array containing documents and their distances to 
            `doc_or_docs`. 
        
        :See Also: :meth:`vsm.viewer.wrappers.dist_doc_doc`
        """
        return dist_doc_doc(doc_or_docs, self.corpus, self.model.context_type,
                              self.model.doc_matrix, weights=weights,
                              print_len=print_len, filter_nan=filter_nan, 
                              label_fn=label_fn, as_strings=True,
                              dist_fn=dist_fn, order=order)
    
    @deprecated_meth("dist_word_doc")
    def sim_word_doc(self, word_or_words, weights=[], label_fn=def_label_fn, 
                      filter_nan=True, print_len=10, as_strings=True, 
                      dist_fn=angle, order='i'):
        pass
    
    def dist_word_doc(self, word_or_words, weights=[], label_fn=def_label_fn, 
                      filter_nan=True, print_len=10, as_strings=True, 
                      dist_fn=angle, order='i'):
        """
        Computes and sorts distances between a word or a list of words to
        every document.
        
        :param word_or_words: Query word(s) to which a pseudo-document is
            created for computation of distances.
        :type word_or_words: string/integer or list of strings/integers
        
        :param weights: Specify weights for each query doc in `word_or_words`. 
            Default uses equal weights (i.e. arithmetic mean)
        :type weights: list of floating point, optional
        
        :param print_len: Number of documents to be displayed. Default is 10.
        :type print_len: int, optional

        :param filter_nan: If `True` not a number entries are filtered.
            Default is `True`.
        :type filter_nan: boolean, optional
 
        :param label_fn: A function that defines how documents are represented.
            Default is :meth:`def_label_fn` which retrieves the labels 
            from corpus metadata.
        :type label_fn: string, optional
        
        :param as_strings: If `True`, returns a list of documents as strings
            rather than indices. Default is `True`.
        :type as_strings: boolean, optional

        :param dist_fn: A distance function from functions in vsm.spatial.
            Default is :meth:`angle`.
        :type dist_fn: string, optional
         
        :param order: Order of sorting 'i' for increasing and 'd' for
            decreasing order. Default is 'i'.
        :type order: string, optional
       
        :returns: an instance of :class:`LabeledColumn`.
            A 2-dim array containing documents and their distances to 
            `word_or_words`. 

        :See Also: :meth:`vsm.viewer.wrappers.dist_word_doc`
        """
        # Resolve `word_or_words`
        if isstr(word_or_words):
            word_or_words = [word_or_words]
        words, labels = list(zip(*[res_word_type(self.corpus, w) for w in word_or_words]))
        words, labels = list(words), list(labels)

        # Generate pseudo-document
        doc = np.zeros((self.model.word_matrix.shape[0],1), dtype=np.float)
        if len(weights) == 0:
            doc[words,:] = np.ones(len(words))
        else:
            doc[words,:] = weights

        doc = np.dot(np.dot(np.diag(1 /self.model.eigenvalues),
                            self.model.word_matrix.T), doc)

        # Compute distances
        d_arr = dist_fn(doc.T, self.model.doc_matrix)

        # Label data
        if as_strings:
            md = self.corpus.view_metadata(self.model.context_type)
            docs = label_fn(md)
            d_arr = enum_sort(d_arr, indices=docs, field_name='doc')
        else:
            d_arr = enum_sort(d_arr, filter_nan=filter_nan)

        if order=='d':
            pass
        elif order=='i':
            d_arr = d_arr[::-1]
        else:
            raise Exception('Invalid order parameter.')

        d_arr = d_arr.view(LabeledColumn)
        # TODO: Finish this header
        d_arr.col_header = 'Words: '
        d_arr.subcol_headers = ['Document', 'Distance']
        d_arr.col_len = print_len
        
        return d_arr

    @deprecated_meth("dismat_word")
    def simmat_words(self, word_list, dist_fn=angle):
        pass
    
    def dismat_word(self, word_list, dist_fn=angle):
        """
        Calculates a distance matrix for a given list of words.

        :param word_list: A list of words whose similarity matrix is to be
            computed.
        :type word_list: list

        :param dist_fn: A distance function from functions in vsm.spatial. 
            Default is :meth:`angle`.
        :type dist_fn: string, optional

        :returns: an instance of :class:`IndexedSymmArray`.
            n x n matrix containing floats where n is the number 
            of words in `word_list`.
        
        :See Also: :meth:`vsm.viewer.wrappers.dismat_word`
        """
        return dismat_word(word_list, self.corpus, 
                             self.model.word_matrix.T, dist_fn=dist_fn)

    @deprecated_meth("dismat_doc")
    def simmat_docs(self, doc_list, dist_fn=angle):
        pass
    
    def dismat_doc(self, doc_list, dist_fn=angle):
        """
        Calculates a distance matrix for a given list of documents.

        :param doc_list: A list of documents whose distance matrix is 
            to be computed.
        :type docs: list, optional
        
        :param dist_fn: A distance function from functions in vsm.spatial. 
            Default is :meth:`angle`.
        :type dist_fn: string, optional

        :returns: an instance of :class:`IndexedSymmArray`.
            n x n matrix containing floats where n is the number 
            of documents. 

        :See Also: :meth:`vsm.viewer.wrappers.dismat_doc`
        """
        return dismat_doc(doc_list, self.corpus, self.model.context_type, 
                            self.model.doc_matrix, dist_fn=dist_fn)




from __future__ import division
from __future__ import print_function
from builtins import zip
from builtins import map
from builtins import range
from past.utils import old_div
import numpy as np
from vsm.zipfile import use_czipfile


__all__ = [ 'init_priors', 'compute_top_doc', 'compute_word_top', 
            'compute_log_prob', 'load_lda', 'save_lda' ]



def init_priors(V=0, K=0, beta=[], alpha=[]):
    # Topic and context priors; set defaults if need be
    if len(beta) > 0:
        beta = (np.array(beta, dtype=np.float).reshape(len(beta), 1))
    else:
        # Default is a flat prior of .01
        beta = np.ones((V, 1), dtype=np.float) * .01

    if len(alpha) > 0:
        alpha = (np.array(alpha, dtype=np.float).reshape(len(alpha), 1))
    else:
        # Default is a flat prior of .01
        alpha = np.ones((K, 1), dtype=np.float) * .01

    return beta, alpha

@use_czipfile
def load_lda(filename, ldaclass):
    """
    A static method for loading a saved `ldaclass` model.
    
    :param filename: Name of a saved model to be loaded.
    :type filename: string
    
    :returns: m : `ldaclass` object
    
    :See Also: :class:`numpy.load`
    """
    try:
        import concurrent.futures
        def load_npz(filename, obj):
            zipfile = np.load(filename)
            return zipfile.__getitem__(obj)
    
        print('Loading LDA data from {}'.format(filename))
    
        arrays_in = np.load(filename)
        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            # load context types
            context_type = executor.submit(load_npz, filename, 'context_type')
            K = executor.submit(load_npz, filename, 'K')
            V = executor.submit(load_npz, filename, 'V')
    
            # load priors
            alpha = executor.submit(load_npz, filename, 'alpha')
            beta = executor.submit(load_npz, filename, 'beta')
    
            # load data
            indices = executor.submit(load_npz, filename, 'indices')
            corpus = executor.submit(load_npz, filename, 'corpus')
            dtype = executor.submit(load_npz, filename, 'dtype')
            Z = executor.submit(load_npz, filename, 'Z')
            Ktype = executor.submit(load_npz, filename, 'Ktype')
    
            iteration = executor.submit(load_npz, filename, 'iteration')
            log_probs = executor.submit(load_npz, filename, 'log_probs')
    
            top_doc = executor.submit(load_npz, filename, 'top_doc')
            word_top = executor.submit(load_npz, filename, 'word_top')
            inv_top_sums = executor.submit(load_npz, filename, 'inv_top_sums')
    
            if 'seed' in arrays_in.files:
                seed = executor.submit(load_npz, filename, 'seed')
                mtrand_state = [executor.submit(load_npz, filename, 
                                                'mtrand_state{}'.format(i)) 
                                    for i in range(5)]
    
            elif 'seeds' in arrays_in.files:
                seeds =  executor.submit(load_npz, filename, 'seeds')
                mtrand_states = [executor.submit(load_npz, filename, 
                                                 'mtrand_states{}'.format(i)) 
                                     for i in range(5)]
    
        m = ldaclass(context_type=context_type.result(), 
                     K=K.result(), V=V.result(), 
                     alpha=alpha.result(), beta=beta.result())
    
        m.indices = indices.result()
        m.corpus = corpus.result()
        m.dtype = m.corpus.dtype
    
        m.Z = Z.result()
        m.Ktype = m.Z.dtype
    
        m.word_top = word_top.result()
        m.top_doc = top_doc.result()
        m.inv_top_sums = inv_top_sums.result()
    
        m.iteration = iteration.result()
        m.log_probs = log_probs.result().tolist()
    
        if 'seed' in arrays_in.files:
            m.seed = int(seed.result())
            m._mtrand_state = [s.result() for s in mtrand_state]
            fns = (str, lambda x: x, int, int, float)
            m._mtrand_state = [f(s) for f, s in zip(fns, m._mtrand_state)]
    
        if 'seeds' in arrays_in.files:
            m.seeds = list(map(int, seeds.result()))
            m._mtrand_states = [s.result() for s in mtrand_states]
            fns = (str, lambda x: x, int, int, float)
            m._mtrand_states = [list(map(f, s)) for f, s in zip(fns, m._mtrand_states)]
            m._mtrand_states = list(zip(*m._mtrand_states))
            m.n_proc = len(m.seeds)
        return m
    except:
        return old_load_lda(filename, ldaclass)


def old_load_lda(filename, ldaclass):
    """
    A static method for loading a saved `ldaclass` model.
    
    :param filename: Name of a saved model to be loaded.
    :type filename: string
    
    :returns: m : `ldaclass` object
    
    :See Also: :class:`numpy.load`
    """
    print('Loading LDA data from {}'.format(filename))
    arrays_in = np.load(filename)

    context_type = arrays_in['context_type'][()]
    K = arrays_in['K'][()]

    if 'm_words' in arrays_in:
        V = arrays_in['m_words'][()]
    else:
        V = arrays_in['V'][()]

    if 'ctx_prior' in arrays_in:
        alpha = arrays_in['ctx_prior']
    elif arrays_in['alpha'].size==1:
        alpha = np.ones(K, dtype=np.float64) * arrays_in['alpha'][()]
    else:
        alpha = arrays_in['alpha']

    if 'top_prior' in arrays_in:
        beta = arrays_in['top_prior']
    elif arrays_in['beta'].size==1:
        beta = np.ones(V, dtype=np.float64) * arrays_in['beta'][()]
    else:
        beta = arrays_in['beta']
        
    m = ldaclass(context_type=context_type, K=K, V=V, alpha=alpha, beta=beta)

    if 'W_indices' in arrays_in:
        m.indices = arrays_in['W_indices']
    elif 'contexts' in arrays_in: 
        m.indices = [s.stop for s in arrays_in['contexts']]
    else:
        m.indices = arrays_in['indices']

    if 'W_corpus' in arrays_in:
        m.corpus = arrays_in['W_corpus']
    else:
        m.corpus = arrays_in['corpus']        

    if 'Z_corpus' in arrays_in:
        m.Z = arrays_in['Z_corpus']        
    else:
        m.Z = arrays_in['Z']

    if 'top_word' in arrays_in:
        m.word_top = arrays_in['top_word'].T
    else:
        m.word_top = arrays_in['word_top']

    if 'doc_top' in arrays_in:
        m.top_doc = arrays_in['doc_top'].T
    elif 'top_ctx' in arrays_in:
        m.top_doc = arrays_in['top_ctx']
    else:
        m.top_doc = arrays_in['top_doc']

    if 'sum_word_top' in arrays_in:
        m.inv_top_sums = 1. / arrays_in['sum_word_top']
    elif 'top_norms' in arrays_in:
        m.inv_top_sums = arrays_in['top_norms']
    else:
        m.inv_top_sums = arrays_in['inv_top_sums']

    if 'iterations' in arrays_in:
        m.iteration = arrays_in['iterations'][()]
    else:
        m.iteration = arrays_in['iteration'][()]

    if 'log_prob' in arrays_in:
        m.log_probs = arrays_in['log_prob'].tolist()
    else:
        m.log_probs = arrays_in['log_probs'].tolist()

    if 'seed' in arrays_in:
        m.seed = int(arrays_in['seed'])
        try:
            m._mtrand_state = (str(arrays_in['mtrand_state0']),
                               arrays_in['mtrand_state1'],
                               int(arrays_in['mtrand_state2']),
                               int(arrays_in['mtrand_state3']),
                               float(arrays_in['mtrand_state4']))
        except TypeError:
            raise Exception(arrays_in['mtrand_state0'],
            type(arrays_in['mtrand_state0']))

    if 'seeds' in arrays_in:
        m.seeds = list(map(int, list(arrays_in['seeds'])))
        m._mtrand_states = list(zip(list(map(str, arrays_in['mtrand_states0'])),
                               arrays_in['mtrand_states1'],
                               list(map(int, arrays_in['mtrand_states2'])),
                               list(map(int, arrays_in['mtrand_states3'])),
                               list(map(float, arrays_in['mtrand_states4']))))
        m.n_proc = len(m.seeds)

    return m


def save_lda(m, filename):
    """
    Saves the model in an `.npz` file.
    
    :param filename: Name of file to be saved.
    :type filename: string
    
    :See Also: :class:`numpy.savez`
    """

    arrays_out = dict()

    arrays_out['context_type'] = m.context_type
    if hasattr(m, 'dtype'):
        if m.dtype == np.uint16:
            arrays_out['dtype'] = 'uint16'
        elif m.dtype == np.uint32:
            arrays_out['dtype'] = 'uint32'
    if hasattr(m, 'Ktype'):
        if m.Ktype == np.uint16:
            arrays_out['Ktype'] = 'uint16'
        elif m.Ktype == np.uint32:
            arrays_out['Ktype'] = 'uint32'

    arrays_out['alpha'] = m.alpha
    arrays_out['beta'] = m.beta

    arrays_out['K'] = m.K
    arrays_out['V'] = m.V
    arrays_out['indices'] = m.indices
    arrays_out['corpus'] = m.corpus
    arrays_out['Z'] = m.Z

    arrays_out['iteration'] = m.iteration
    dt = dtype=[('i', np.int), ('v', np.float)]
    arrays_out['log_probs'] = np.array(m.log_probs, dtype=dt)

    potential_views = ['top_doc','word_top','inv_top_sums']
    arrays_out['top_doc'] = m.top_doc
    arrays_out['word_top'] = m.word_top
    arrays_out['inv_top_sums'] = m.inv_top_sums

    if hasattr(m,'seed'):
        arrays_out['seed'] = m.seed
    if hasattr(m,'seeds'):
        arrays_out['seeds'] = m.seeds
    if hasattr(m, '_mtrand_state'):
        for i,s in enumerate(m._mtrand_state):
            key = 'mtrand_state{0}'.format(i)
            arrays_out[key] = s
    
    if hasattr(m, '_mtrand_states'):
        for i,s in enumerate(zip(*m._mtrand_states)):
            key = 'mtrand_states{0}'.format(i)
            arrays_out[key] = s
    
    print('Saving LDA model to {}'.format(filename))
    use_czipfile(np.savez)(filename, **arrays_out)


def compute_log_prob(W, Z, word_top, top_doc):
    log_wt = np.log(word_top / word_top.sum(0))
    log_td = np.log(top_doc / top_doc.sum(0))
    log_prob = 0
    for i in range(len(W)):
        W_i, Z_i = W[i], Z[i]
        for j in range(len(W_i)):
            w, k = W_i[j], Z_i[j]
            log_prob += log_wt[w,k] + log_td[k,i]
    return log_prob


def compute_top_doc(Z, K, alpha=[]):
    """
    Takes a topic assignment Z, the number of topics K and optionally
    a document prior and returns the topic-document matrix which is
    used in the LDA model objects.
    """    
    if len(alpha)==0:
        top_doc = np.zeros((K, len(Z)), dtype=np.float)
    else:
        top_doc = np.zeros((K, len(Z)), dtype=np.float) + alpha

    for i in range(len(Z)):
        for j in range(len(Z[i])):
            z = Z[i][j]
            top_doc[z][i] += 1

    return top_doc


def compute_word_top(W, Z, K, V, beta=[]):
    """
    Takes a list of documents W, a topic assignment Z, the number of
    topics K, the number of words in the vocabulary V and optionally a
    topic prior and returns the word-topic matrix which is used in the
    LDA model objects.
    """    
    if len(beta)==0:
        word_top = np.zeros((V, K), dtype=np.float)
    else:
        word_top = np.zeros((V, K), dtype=np.float) + beta

    for i in range(len(Z)):
        for j in range(len(Z[i])):
            w = W[i][j]
            k = Z[i][j]
            word_top[w][k] += 1

    return word_top



from builtins import object
import multiprocessing as mp
import platform, warnings

import numpy as np
from scipy.sparse import hstack

from vsm.spatial import count_matrix
from vsm.split import *
from vsm.model.base import *


__all__ = ['TF', 'TfSeq', 'TfMulti']



class TfSeq(BaseModel):
    """
    Trains a term-frequency model. 

    In a term-frequency model, the number of occurrences of a word
    type in a context is counted for all word types and documents. Word
    types correspond to matrix rows and documents correspond to matrix
    columns.

    :See Also: :class:`vsm.model.base`, :class:`vsm.corpus.Corpus`,
        :class:`scipy.sparse.coo_matrix`
    """
    
    def __init__(self, corpus=None, context_type=None):
        """
        Initialize TfSeq.

        :param corpus: A Corpus object containing the training data.
        :type corpus: Corpus
    
        :param context_type: A string specifying the type of context over which
            the model trainer is applied.
        :type context_type: string
        """

        self.context_type = context_type
        if corpus:
            self.corpus = corpus.corpus
            self.docs = corpus.view_contexts(context_type, as_slices=True)
            self.V = corpus.words.size
        else:
            self.corpus = []
            self.docs = []
            self.V = 0


    def train(self):
        """
        Counts word-type occurrences per context and stores the results in
        `self.matrix`.
        """
        self.matrix = count_matrix(self.corpus, self.docs, self.V)


    
class TfMulti(TfSeq):
    """
    Trains a term-frequency model. 

    In a term-frequency model, the number of occurrences of a word
    type in a context is counted for all word types and documents. Word
    types correspond to matrix rows and documents correspond to matrix
    columns.

    The data structure is a sparse integer matrix.

    :See Also: :class:`vsm.model.base.BaseModel`, :class:`vsm.corpus.Corpus`,
        :class:`scipy.sparse.coo_matrix`
    """
    def __init__(self, corpus=None, context_type=None):
        """
        Initialize TfMulti.

        :param corpus: A Corpus object containing the training data
        :type corpus: Corpus, optional
    
        :param context_type: A string specifying the type of context over which
            the model trainer is applied.
        :type context_type: string, optional
        """
        self._read_globals = False
        self._write_globals = False

        super(TfMulti, self).__init__(corpus=corpus, context_type=context_type)


    def _move_globals_to_locals(self):
        
        self._write_globals = False
        self.V = self.V
        self.corpus = self.corpus
        self._read_globals = False
        global _V, _corpus
        del _V, _corpus


    def _move_locals_to_globals(self):
        
        self._write_globals = True
        self.V = self.V
        self.corpus = self.corpus
        self._read_globals = True
        del self._V_local, self._corpus_local


    @property
    def corpus(self):
        if self._read_globals:
            return np.frombuffer(_corpus, np.int32)
        return self._corpus_local

    @corpus.setter
    def corpus(self, a):
        if self._write_globals:
            global _corpus
            if not '_corpus' in globals():
                _corpus = mp.Array('i', len(a), lock=False)
            _corpus[:] = a
        else:
            self._corpus_local = a

    @property
    def V(self):
        if self._read_globals:
            return _V.value
        return self._V_local

    @V.setter
    def V(self, V):
        if self._write_globals:
            global _V
            _V = mp.Value('i', V, lock=False)
        else:
            self._V_local = V



    def train(self, n_proc=2):
        """
        Takes a number of processes `n_proc` over which to map and reduce.

        :param n_procs: Number of processors.
        :type n_procs: int
        """
        self._move_locals_to_globals()

        doc_indices = mp_split_ls(self.docs, n_proc)

        p=mp.Pool(n_proc)
        cnt_mats = p.map(tf_fn, doc_indices)
        p.close()

        self.matrix = hstack(cnt_mats, format='coo')

        self._move_globals_to_locals()



def tf_fn(ctx_sbls):
    """
    The map function for vsm.model.TfMulti. Takes a list of documents
    as slices and returns a count matrix.
    
    :param ctx_sbls: list of documents as slices.
    :type ctx_sbls: list of slices

    :returns: a count matrix
    """
    offset = ctx_sbls[0].start
    corpus = _corpus[offset: ctx_sbls[-1].stop]
    slices = [slice(s.start-offset, s.stop-offset) for s in ctx_sbls]
    return count_matrix(corpus, slices, _V.value)


class TF(object):
    """
    Depending on the boolean parameter `multiprocessing`, returns and
    initializes an instance of either TfSeq or TfMulti.

    Note that on Windows platforms, `multiprocessing` is not implemented.
    In contrast to LdaCgsMulti, LDA always returns a valid object. Instead
    of raising a NotImplementedError, LDA issues a RuntimeWarning, notifying 
    the user the sequental algorithm is being used.
    """
    def __new__(cls, corpus=None, context_type=None, multiprocessing=False):

        kwargs = dict(corpus=corpus, context_type=context_type)
        
        if multiprocessing and platform.system() != 'Windows':
            return TfMulti(**kwargs)
        else:
            if platform.system() == 'Windows':
                warnings.warn("""Multiprocessing is not implemented on Windows.
                Defaulting to sequential algorithm.""", RuntimeWarning)
            return TfSeq(**kwargs)



"""
[General documentation about the :mod:`exceptions` submodule]
"""
from __future__ import absolute_import

from . import compatibility
from .compatibility import *

__all__ = compatibility.__all__



from __future__ import print_function
from builtins import str
from builtins import range
from builtins import object
import os

import numpy as np

from vsm.structarr import arr_add_field
from vsm.split import split_corpus


__all__ = [ 'BaseCorpus', 'Corpus', 'add_metadata',
            'align_corpora','binary_search' ]

from bisect import bisect_left
from datetime import datetime
from vsm.zipfile import use_czipfile
from copy import deepcopy



def binary_search(a, x, lo=0, hi=None):   # can't use a to specify default for hi
    hi = hi if hi is not None else len(a) # hi defaults to len(a)   
    pos = bisect_left(a,x,lo,hi)          # find insertion position
    return (pos if pos != hi and a[pos] == x else -1) # don't walk off the end
"""
def binary_search_set(a,x):
    pos = a.bisect_left(x)
    return (pos if pos != len(a)  and a[pos] == x else -1) # don't walk off the end
"""

def load_npz(filename, obj):
    zipfile = np.load(filename)
    return zipfile.__getitem__(obj)

class BaseCorpus(object):
    """
    A BaseCorpus object stores a corpus along with its tokenizations
    (e.g., as sentences, paragraphs or documents).

    BaseCorpus aims to provide an efficient method for viewing the
    corpus in tokenized form (i.e., without copying). It currently
    achieves this by storing the corpus as a numpy `ndarray`. Viewing
    a tokenization is carried out using the `view` facilities of numpy
    ndarrays. See documentation on `numpy.ndarray` for further
    details.

    :param corpus: Array, typically of strings or integers, of atomic words
        (or tokens) making up the corpus.
    :type corpus: array-like

    :param dtype: The data-type used to interpret the corpus. If omitted, the
        data-type is determined by `numpy.asarray`. Default is `None`.
    :type dtype: data-type, optional

    :param context_data: Each element in `context_data` is an array 
        containing the indices marking the context boundaries. 
        An element in `context_data` is intended for use as a value for 
        the `indices_or_sections` parameter in `numpy.split`. 
        Elements of `context_data` may also be 1-D arrays whose elements
        are pairs, where the first element is a context boundary and 
        the second element is metadata associated with that context preceding
        that boundary. For example, (250, 'dogs') might indicate that 
        the 'article' context ending at the 250th word of the corpus is named 
        'dogs'. Default is `None`.
    :type context_data:  list with 1-D array-like elements, optional

    :param context_types: Each element in `context_types` is a type of a
        tokenization in `context_data`.
    :type context_types: array-like, optional

    :param remove_empty: If True, empty tokenizations are removed. Default is
        `True`.
    :type remove_empty: boolean, optional

    :param to_array: If True, converts all values to a numpy array. If False,
        data is left in input format. False is used when `BaseCorpus` is used
        in a constructor for an advanced tokenization like `Corpus`. Default is
        `True`.
    :type to_array: boolean, optional

    :attributes: 
        * **corpus**  (1-dimensional array)
            Stores the value of the `corpus` parameter after it has been cast
            to an array of data-type `dtype` (if provided).
        * **words**  (1-dimensional array)
            The indexed set of atomic words appearing in `corpus`.
        * **context_types**  (1-dimensional array-like)
        * **context_data**   (list of 1-D array-like)

    :methods: 
        * **meta_int**
            Takes a type of tokenization and a query and returns the index of
            the metadata found in the query.
        * **get_metadatum**
            Takes a type of tokenization and a query and returns the metadatum
            corresponding to the query and the field.
        * **view_contexts**
            Takes a type of tokenization and returns a view of the corpus
            tokenized accordingly.
        * **view_metadata**
            Takes a type of tokenization and returns a view of the metadata
            of the tokenization.
        * **tolist**
            Returns Corpus object as a list of lists.
        * **remove_empty**
            Removes empty documents in the corpus.

    :See Also: :class:`Corpus`

    **Examples**

    >>> corpus = ['the', 'dog', 'chased', 'the', 'cat',
                  'the', 'cat', 'ran', 'away']
    >>> context_types = ['sentences']
    >>> context_data = [np.array([(5, 'transitive'), (9, 'intransitive')],
                        dtype=[('idx', '<i8'), ('sent_label', '|S16')])]

    >>> from vsm.corpus import BaseCorpus
    >>> c = BaseCorpus(corpus, context_types=context_types, context_data=context_data)
    >>> c.corpus
    array(['the', 'dog', 'chased', 'the', 'cat', 'the', 'cat',
           'ran', 'away'], dtype='|S6')
           
    >>> c.words
    array(['ran', 'away', 'chased', 'dog', 'cat', 'the'],
          dtype='|S6')

    >>> c.meta_int('sentences',{'sent_label': 'intransitive'})
    1

    >>> b.get_metadatum('sentences', {'sent_label': 'intransitive'}, 'sent_label') 
    'intransitive'

    >>> c.view_contexts('sentences')
    [array(['the', 'dog', 'chased', 'the', 'cat'],
          dtype='|S6'),
     array(['the', 'cat', 'ran', 'away'],
          dtype='|S6')]

    >>> c.view_metadata('sentences')[0]['sent_label']
    'transitive'
    """
    __slots__ = ['words', 'context_data', 'corpus','dtype','stopped_words',
                 'context_types']
    def __init__(self,
                 corpus,
                 dtype=None,
                 context_types=[],
                 context_data=[],
                 remove_empty=False,
                 to_array=True):

        if to_array:
            self.corpus = np.asarray(corpus, dtype=dtype)
            self.dtype = self.corpus.dtype
        else:
            self.corpus = corpus[:]
            self.dtype = dtype

        # Since np.unique attempts to make a whole contiguous copy of the
        # corpus array, we instead use a sorted set and cast to a np array
        # equivalent to self.words = np.unique(self.corpus)
        self.words = np.asarray(sorted(set(self.corpus)), dtype=np.object_)

        self.context_data = []
        for t in context_data:
            if self._validate_indices(t['idx']):
                self.context_data.append(t)
        
        self._gen_context_types(context_types)
        
        self.stopped_words = set()

        if remove_empty:
            self.remove_empty()

    def __len__(self):
        """
        Returns the number of tokens in the corpus.

        :See Also: `len(self.words)` for the number of unique tokens.
        """

        return len(self.corpus)

    def _gen_context_types(self, context_types):
        """
        Missing context types are filled in with 'ctx_' + an index.
        """
        if self.context_data:
            a = len(context_types) if context_types else 0
            for i in range(a, len(self.context_data)):
                context_types.append('ctx_' + str(i))

        self.context_types = context_types


    def _validate_indices(self, indices):
        """
        Checks for invalid tokenizations. Specifically, checks to see
        that the list of indices are sorted and are in range. Ignores
        empty tokens.

        :param indices: 
        :type indices : 1-D integer array-like

        :returns: `True` if the indices are validated

        :raises: Exception

        :See Also: :class:`BaseCorpus`
        """
        for i, j in enumerate(indices):
            if i < len(indices) - 1 and j > indices[i + 1]:
                msg = 'malsorted tokenization:'\
                      ' ctx ' + str(j) + ' and ' + str(indices[i + 1])

                raise Exception(msg)
                    
            if j > len(self.corpus):
                msg = 'invalid tokenization'\
                      ' : ' + str(j) + ' is out of range ('\
                      + str(len(self.corpus)) + ')'
                
                raise Exception(msg)

        return True

    
    def remove_empty(self):
        """
        Removes empty tokenizations, if `Corpus` object is not empty.
        """    
        if self:
            for j, t in enumerate(self.context_types):
                token_list = self.view_contexts(t)
                
                indices = np.array([ctx.size != 0 for ctx in token_list], dtype=np.bool)
                self.context_data[j] = self.context_data[j][indices]


    def view_metadata(self, ctx_type):
        """
        Displays the metadata corresponding to a tokenization of the
        corpus. This method can be used in :class:`Corpus` as well as
        :class:`BaseCorpus`

        :param ctx_type: The type of a tokenization.
        :type ctx_type: string-like

        :returns: The metadata for a tokenization.

        :See Also: :class:`BaseCorpus`, :class:`Corpus`
        """
        i = self.context_types.index(ctx_type)
        return self.context_data[i]


    def meta_int(self, ctx_type, query):
        """
        Returns the index of the metadata found in the query.

        :param ctx_type: The type of a tokenization.
        :type ctx_type: string-like
        
        :param query: Dictionary with a key, value being a field, label
            in metadata.
        :type query: dictionary-like

        :returns: The index of the metadata found in the query.
        
        :raises: KeyError

        :See Also: :class:`BaseCorpus`
        """

        tok = self.view_metadata(ctx_type)

        ind_set = np.ones(tok.size, dtype=bool)
        for k,v in query.items():
            try:
                ind_set = np.logical_and(ind_set, (tok[k] == v))
            except UnicodeDecodeError:
                v = v.decode('utf-8')
                ind_set = np.logical_and(ind_set, (tok[k] == v))

        n = np.count_nonzero(ind_set)
        if n == 0:
            raise KeyError('No token fits the description: ' +
                ', '.join(['{q}:{l}'.format(q=k, l=v) 
                                for k,v in query.items()]))
        elif n > 1:
            msg = ('Multiple tokens fit that description:\n'
                   + str(tok[ind_set]))
            raise KeyError(msg)

        return ind_set.nonzero()[0][0]


    def get_metadatum(self, ctx_type, query, field):
        """
        Returns the metadatum corresponding to the query and the field.

        :param ctx_type: The type of a tokenization.
        :type ctx_type: string-like
        
        :param query: Dictionary with a key, value being a field, label
            in metadata.
        :type query: dictionary-like
        
        :param field: Field of the metadata
        :type field: string

        :returns: The metadatum corresponding to the query and the field.

        :See Also: :class:`BaseCorpus`
        """
        i = self.meta_int(ctx_type, query)
        return self.view_metadata(ctx_type)[i][field]



    def view_contexts(self, ctx_type, as_slices=False, as_indices=False):
        """
        Displays a tokenization of the corpus.

        :param ctx_type: The type of a tokenization.
        :type ctx_type: string-like

        :param as_slices: If True, a list of slices corresponding to 
            'ctx_type' is returned. Otherwise, integer representations
            are returned. Default is `False`.
        :type as_slices: Boolean, optional

        :Returns: A tokenized view of `corpus`.

        :See Also: :class:`BaseCorpus`, :meth:`numpy.split`
        """
        indices = self.view_metadata(ctx_type)['idx']

        if as_indices:
            return indices
        
        if as_slices:
            if len(indices) == 0:
                return [slice(0, 0)]
                
            slices = []
            slices.append(slice(0, indices[0]))
            for i in range(len(indices) - 1):
                slices.append(slice(indices[i], indices[i+1]))
            return slices        
            
        return split_corpus(self.corpus, indices)


    def tolist(self, context_type):
        """
        Returns Corpus object as a list of lists.
        """
        return self.view_contexts(context_type)


    
class Corpus(BaseCorpus):
    """
    The goal of the Corpus class is to provide an efficient representation\
    of a textual corpus.

    A Corpus object contains an integer representation of the text and
    maps to permit conversion between integer and string
    representations of a given word.

    As a BaseCorpus object, it includes a dictionary of tokenizations
    of the corpus and a method for viewing (without copying) these
    tokenizations. This dictionary also stores metadata (e.g.,
    document names) associated with the available tokenizations.

    :param corpus: A string array representing the corpus as a sequence of
        atomic words.
    :type corpus: array-like

    :param context_data: Each element in `context_data` is an array containing 
        the indices marking the token boundaries. An element in `context_data` is
        intended for use as a value for the `indices_or_sections`
        parameter in `numpy.split`. Elements of `context_data` may also be
        1-D arrays whose elements are pairs, where the first element
        is a context boundary and the second element is metadata
        associated with that context preceding that boundary. For
        example, (250, 'dogs') might indicate that the 'article' context
        ending at the 250th word of the corpus is named 'dogs'.
        Default is `None`.
    :type context_data: list-like with 1-D integer array-like elements, optional
    
    :param context_types: Each element in `context_types` is a type of a context
        in `context_data`.
    :type context_types: array-like, optional

    :attributes: 
        * **corpus** (1-D 32-bit integer array)
            corpus is the integer representation of the input string array-like
            value of the corpus parameter
        * **words** (1-D string array)
            The indexed set of strings occurring in corpus. It is a string-typed array.
        * **words_in** (1-D 32-bit integer dictionary)
            A dictionary whose keys are `words` and whose values are their 
            corresponding integers (i.e., indices in `words`).
        
    :methods:
        * **view_metadata**
            Takes a type of tokenization and returns a view of the metadata
            of the tokenization.
        * **view_contexts**
            Takes a type of tokenization and returns a view of the corpus tokenized
            accordingly. The optional parameter `strings` takes a boolean value: 
            True to view string representations of words; False to view integer 
            representations of words. Default is `False`.
        * **save**
            Takes a filename and saves the data contained in a Corpus object to 
            a `npy` file using `numpy.savez`.
        * **load**
            Static method. Takes a filename, loads the file data into a Corpus
            object and returns the object.
        * **apply_stoplist**
            Takes a list of stopwords and returns a copy of the corpus with 
            the stopwords removed.
        * **tolist**
            Returns Corpus object as a list of lists of either integers or strings, 
            according to `as_strings`.
        
    :See Also: :class:`BaseCorpus`

    **Examples**

    >>> text = ['I', 'came', 'I', 'saw', 'I', 'conquered']
    >>> context_types = ['sentences']
    >>> context_data = [np.array([(2, 'Veni'), (4, 'Vidi'), (6, 'Vici')],
                            dtype=[('idx', '<i8'), ('sent_label', '|S6')])]

    >>> from vsm.corpus import Corpus
    >>> c = Corpus(text, context_types=context_types, context_data=context_data)
    >>> c.corpus
    array([0, 1, 0, 2, 0, 3], dtype=int32)
    
    >>> c.words
    array(['I', 'came', 'saw', 'conquered'],
          dtype='|S9')

    >>> c.words_int['saw']
    2

    >>> c.view_contexts('sentences')
    [array([0, 3], dtype=int32), array([0, 2], dtype=int32),
     array([0, 1], dtype=int32)]

    >>> c.view_contexts('sentences', as_strings=True)
    [array(['I', 'came'], 
          dtype='|S9'),
     array(['I', 'saw'], 
          dtype='|S9'),
     array(['I', 'conquered'], 
          dtype='|S9')]

    >>> c.view_metadata('sentences')[1]['sent_label']
    'Vidi'
    
    >>> c = c.apply_stoplist(['saw'])
    >>> c.words
    array(['I', 'came', 'conquered'], 
      dtype='|S9')

    """
    __slots__ = ['words', 'context_data', 'corpus','dtype','stopped_words',
                 'words_int', 'context_types']

    def __init__(self,
                 corpus,
                 context_types=[],
                 context_data=[],
                 remove_empty=True):

        super(Corpus, self).__init__(corpus,
                                     context_types=context_types,
                                     context_data=context_data,
                                     dtype=np.object_,
                                     remove_empty=False,
                                     to_array=False)

        self._set_words_int()

        # Integer encoding of a string-type corpus
        if len(self.words) < 2 ** 16:
            self.dtype = np.uint16
        else:
            self.dtype = np.uint32

        self.corpus = np.asarray([self.words_int[word]
                                  for word in self.corpus 
                                      if str(word) not in ['\x00']
                                      ],
                                 dtype=self.dtype)

        self.stopped_words = set()

        if remove_empty:
            self.remove_empty()



    def _set_words_int(self):
        """
        Mapping of words to their integer representations.
        """
        self.words_int = dict((t,i) for i,t in enumerate(self.words))


    def view_contexts(self, ctx_type, as_strings=False, as_slices=False, as_indices=False):
        """
        Displays a tokenization of the corpus.

        :param ctx_type: The type of a tokenization.
        :type ctx_type: string-like

        :param as_strings: If True, string representations of words are returned.
            Otherwise, integer representations are returned. Default
            is `False`.
        :type as_strings: Boolean, optional

        :param as_slices: If True, a list of slices corresponding to 'ctx_type'
            is returned. Otherwise, integer representations are returned.
            Default is `False`.
        :type as_slices: Boolean, optional

        :returns: A tokenized view of `corpus`.

        :See Also: :class:`Corpus`, :class:`BaseCorpus`
        """     
        if as_strings:
            token_list = super(Corpus, self).view_contexts(ctx_type)
            token_list_ = []
            for token in token_list:
                token = self.words[token]
                token_list_.append(token)

            return token_list_

        return super(Corpus, self).view_contexts(ctx_type,
                                                 as_slices=as_slices,
                                                 as_indices=as_indices)
    
    
    def tolist(self, context_type, as_strings=False):
        """
        Returns Corpus object as a list of lists of either integers or
        strings, according to `as_strings`.

        :param context_type: The type of tokenization.
        :type context_type: string

        :param as_strings: If True, string representations of words are returned.
            Otherwise, integer representations are returned. Default
            is `False`.
        :type as_strings: Boolean, optional
        
        :returns: List of lists
        """
        ls = self.view_contexts(context_type, as_strings=as_strings)
        return [arr.tolist() for arr in ls]

    @staticmethod
    def load(file=None, corpus_dir=None,
             corpus_file='corpus.npy',
             words_file='words.npy',
             metadata_file='metadata.npy'):
        """Loads data into a Corpus object. 
        
        :param file: The file to read. See `numpy.load` for further
            details. Assumes file has been constructed as by
            `Corpus.save`. This option is exclusive of `corpus_dir`.
        :type file: str-like or file object

        :param corpus_dir: A directory containing the files
        `corpus_file`, `words_file`, `metadata_file`, from which to
        instantiate a Corpus object. This option is ignored if `file`
        is not `None`.
        :type corpus_dir: string

        :param corpus_file: File under `corpus_dir` containing the
        corpus data, stored as a numpy array of integers in an `npy`
        file.
        :type corpus_file: string or file object
        
        :param words_file: File under `corpus_dir` containing the
        corpus vocabulary, stored as a numpy array of strings in an
        `npy` file.  
        :type words_file: string or file object
        
        :param metadata_file: File under `corpus_dir` containing the
        corpus metadata, stored as a numpy stuctured array in an `npy`
        file. Note that this structured array should contain a file
        `idx` which stores the integer indices marking the document
        boundaries.
        :type corpus_file: string or file object
        
        :returns: A Corpus object.

        :See Also: :class:`Corpus`, :meth:`Corpus.save`, :meth:`numpy.load`

        """


        try:
            __IPYTHON__
            notebook = True
            print("Running from notebook, using serial load function.")
        except NameError:
            notebook = False


        if not notebook and file is not None:
            return Corpus._parallel_load(file)

        elif notebook and file is not None:
            return Corpus._serial_load(file)

        elif not corpus_dir is None:
            return Corpus._multifile_load(corpus_dir=corpus_dir,
                corpus_file=corpus_file, words_file=words_file,
                metadata_file=metadata_file)


    @staticmethod
    @use_czipfile
    def _parallel_load(file):
        import concurrent.futures
        import functools
        from pickle import PickleError, UnpicklingError

        c = Corpus([], remove_empty=False)
        # submit futures
        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            c.corpus = executor.submit(load_npz, file, 'corpus')
            c.words =  executor.submit(load_npz, file, 'words')

            c.context_types = executor.submit(load_npz, file, 'context_types')
            c.stopped_words = executor.submit(load_npz, file, 'stopped_words')

            c.dtype = executor.submit(load_npz, file, 'dtype')
            concurrent.futures.wait([c.context_types])
            c.context_types = c.context_types.result().tolist()
            c.context_data = ['context_data_{}'.format(n) for n in c.context_types]
            c.context_data = [executor.submit(load_npz, file, name) for name in c.context_data]


        c.corpus = c.corpus.result()
        c.words = c.words.result()
        try:
            c.dtype = c.dtype.result() 
        except KeyError:
            c.dtype = c.corpus.dtype 

        c.context_data = [future.result() for future in c.context_data]

        try:
            c.stopped_words = set(c.stopped_words.result().tolist())
        except:
            c.stopped_words = set()

        # reset words_int dictionary
        c._set_words_int()

        return c

    @staticmethod
    def _serial_load(file):
        arrays_in = np.load(file)

        c = Corpus([], remove_empty=False)
        c.corpus = arrays_in['corpus']
        c.words = arrays_in['words']
        c.context_types = arrays_in['context_types'].tolist()
        try:
            c.stopped_words = set(arrays_in['stopped_words'].tolist())
        except:
            c.stopped_words = set()

        c.context_data = list()
        for n in c.context_types:
            t = arrays_in['context_data_' + n]
            c.context_data.append(t)

        c._set_words_int()

        return c

    @staticmethod
    def _multifile_load(corpus_dir=None,
                        corpus_file='corpus.npy',
                        words_file='words.npy',
                        metadata_file='metadata.npy'):
        c = Corpus([], remove_empty=False)

        c.corpus = np.load(os.path.join(corpus_dir, corpus_file))
        c.words = np.load(os.path.join(corpus_dir, words_file))
        c._set_words_int()
        c.context_types = [ 'document' ]
        c.context_data = [ np.load(os.path.join(corpus_dir, metadata_file)) ]

        return c

    @use_czipfile
    def save(self, file):
        """
        Saves data from a Corpus object as an `npz` file.
        
        :param file: Designates the file to which to save data. See
            `numpy.savez` for further details.
        :type file: str-like or file-like object
            
        :returns: None

        :See Also: :class:`Corpus`, :meth:`Corpus.load`, :meth:`np.savez`
        """
        import vsm.zipfile
        print('Saving corpus as {}'.format(file))
        arrays_out = dict()
        arrays_out['corpus'] = self.corpus
        arrays_out['words'] = self.words
        arrays_out['context_types'] = np.asarray(self.context_types)
        arrays_out['stopped_words'] = np.asarray(self.stopped_words)
        arrays_out['dtype'] = str(self.dtype)

        for i,t in enumerate(self.context_data):
            key = 'context_data_' + self.context_types[i]
            arrays_out[key] = t

        np.savez(file, **arrays_out)


    def in_place_stoplist(self, stoplist=None, freq=0):
        """ 
        Changes a Corpus object with words in the stoplist removed and with 
        words of frequency <= `freq` removed.
        
        :param stoplist: The list of words to be removed.
        :type stoplist: list

        :param freq: A threshold where words of frequency <= 'freq' are
            removed. Default is 0.
        :type freq: integer, optional
            
        :returns: Copy of corpus with words in the stoplist and words
            of frequnecy <= 'freq' removed.

        :See Also: :class:`Corpus`
        """
        from sortedcontainers import SortedSet, SortedList
        stop = SortedSet()

        if stoplist:
            for t in stoplist:
                if t in self.words_int:
                    stop.add(self.words_int[t])

        if freq:
            cfs = np.bincount(self.corpus)
            freq_stop = np.where(cfs <= freq)[0]
            stop.update(freq_stop)


        if not stop:
            # print 'Stop list is empty.'
            return self
    
        # print 'Removing stop words', datetime.now()
        f = np.vectorize(stop.__contains__)

        # print 'Rebuilding context data', datetime.now()
        context_data = []

        BASE = len(self.context_data) - 1
        # gathering list of new indicies from narrowest tokenization
        def find_new_indexes(INTO, BASE=-1):
            locs = np.where(np.in1d(self.context_data[BASE]['idx'], self.context_data[INTO]['idx']))[0]

            # creating a list of lcoations that are non-identical
            new_locs = np.array([loc for i, loc in enumerate(locs)
                                     if i+1 == len(locs) or self.context_data[BASE]['idx'][locs[i]] != self.context_data[BASE]['idx'][locs[i+1]]])

            # creating a search for locations that ARE identical
            idxs = np.insert(self.context_data[INTO]['idx'], [0,-1], [-1,-1])
            same_spots = np.where(np.equal(idxs[:-1], idxs[1:]))[0]

            # readding the identical locations
            really_new_locs = np.insert(new_locs, same_spots, new_locs[same_spots-1])
            return really_new_locs

        # Calculate new base tokens
        tokens = self.view_contexts(self.context_types[BASE])
        new_corpus = []
        spans = []
        for t in tokens:
            new_t = t[np.logical_not(f(t))] if t.size else t
            
            # TODO: append to new_corpus as well
            spans.append(new_t.size if new_t.size else 0)
            if new_t.size:
                new_corpus.append(new_t)

        # Stopped all words from Corpus
        if not new_corpus:
            return Corpus([])

        new_base = self.context_data[BASE].copy()
        new_base['idx'] = np.cumsum(spans)

        context_data = []
        # calculate new tokenizations for every context_type
        for i in range(len(self.context_data)):
            if i == BASE:
                context_data.append(new_base)
            else:
                context = self.context_data[i].copy()
                context['idx'] = new_base['idx'][find_new_indexes(i, BASE)]
                context_data.append(context)

        del self.context_data
        self.context_data = context_data

        # print 'Rebuilding corpus and updating stop words', datetime.now()
        self.corpus = np.concatenate(new_corpus)
        #self.corpus[f(self.corpus)]
        self.stopped_words.update(self.words[stop])

        #print 'adjusting words list', datetime.now()
        new_words = np.delete(self.words, stop)

        # print 'rebuilding word dictionary', datetime.now()
        new_words_int = dict((word,i) for i, word in enumerate(new_words)) 
        old_to_new =  dict((self.words_int[word],i) for i, word in enumerate(new_words)) 

        #print "remapping corpus", datetime.now()
        f = np.vectorize(old_to_new.__getitem__)
        self.corpus[:] = f(self.corpus)

        #print 'storing new word dicts', datetime.now()
        self.words = new_words
        self.words_int = new_words_int

        return self

    def apply_stoplist(self, stoplist=[], freq=0):
        """ 
        Takes a Corpus object and returns a copy of it with words in the
        stoplist removed and with words of frequency <= `freq` removed.
        
        :param stoplist: The list of words to be removed.
        :type stoplist: list

        :param freq: A threshold where words of frequency <= 'freq' are
            removed. Default is 0.
        :type freq: integer, optional
            
        :returns: Copy of corpus with words in the stoplist and words
            of frequnecy <= 'freq' removed.

        :See Also: :class:`Corpus`
        """
        new_c = deepcopy(self)
        return new_c.in_place_stoplist(stoplist=stoplist, freq=freq)

    def __deepcopy__(self, memo):
        _c = type(self)([], remove_empty=False)
        _c.corpus = np.copy(self.corpus)
        _c.words = np.copy(self.words)
        _c.context_types = self.context_types[:]
        _c.stopped_words = deepcopy(self.stopped_words, memo)
        _c.dtype = self.dtype
        _c.words_int = deepcopy(self.words_int, memo)
        _c.context_data = [np.copy(ctx) for ctx in self.context_data]

        return _c



def add_metadata(corpus, ctx_type, new_field, metadata):
    """
    Returns a corpus with metadata added.

    :param corpus: Corpus object to add new metadata to.
    :type corpus: :class:`Corpus`

    :param ctx_type: A type of tokenization.
    :type ctx_type: string

    :param new_field: Field name of the new metadata.
    :type new_field: string

    :param metadata: List of values to be added to `corpus`.
    :type metdata: list

    :returns: Corpus with new metadata added to the existing metdata.

    :See Also: :class:`Corpus`
    """
    i = corpus.context_types.index(ctx_type)
    md = corpus.context_data[i]
    corpus.context_data[i] = arr_add_field(md, new_field, metadata)

    return corpus



def align_corpora(old_corpus, new_corpus, remove_empty=True):
    """Takes two Corpus objects `old_corpus` and `new_corpus` and returns
    a copy of `new_corpus` with the following modifications: (1) the
    word to integer mapping agrees with that of `old_corpus` and (2)
    words in `new_corpus` which do not appear in `old_corpus` are
    removed from the corpus. Empty documents are removed.

    """
    new_words = [w for w in new_corpus.words if w not in old_corpus.words]
    out = new_corpus.apply_stoplist(new_words)
    if remove_empty:
        out.remove_empty()

    int_words = out.words
    words_int = old_corpus.words_int
    int_int = {}
    for i in range(len(int_words)):
        int_int[i] = words_int[int_words[i]]

    for i in range(len(out.corpus)):
        out.corpus[i] = int_int[out.corpus[i]]
    out.words = old_corpus.words.copy()
    out._set_words_int()

    return out




"""
[General documentation about the :mod:`viewer` submodule]
"""


from . import beagleviewer
from .beagleviewer import *
from . import ldacgsviewer
from .ldacgsviewer import *
from . import lsaviewer
from .lsaviewer import *
from . import tfviewer
from .tfviewer import *
from . import tfidfviewer
from .tfidfviewer import *

__all__ = beagleviewer.__all__[:]
__all__ += ldacgsviewer.__all__
__all__ += lsaviewer.__all__
__all__ += tfviewer.__all__
__all__ += tfidfviewer.__all__




import numpy as np

from scipy.sparse import issparse, csr_matrix, coo_matrix

from vsm.spatial import angle, JS_dist
from vsm.structarr import *
from vsm.viewer.types import *
from vsm.viewer.labeleddata import *


__all__ = ['def_label_fn', 'doc_label_name',
           'dismat_doc', 'dismat_top', 'dismat_word', 'dist_doc_doc',
           'dist_word_doc', 'dist_word_top', 'dist_word_word', 
           'dist_top_doc', 'dist_top_top']

# 
# A default function for constructing document labels from corpus
# metadata
#


def def_label_fn(metadata):
    """
    Takes metadata and returns an array of strings that are 
    strings constructed with all "labels" in the metdata.

    :param metadata: Metadata most likely retrieved from
        Corpus.view_metadata(ctx_type).
    :type metadata: array

    :returns: Array of strings. Each entry is a string with different
        metadata separated by ', '. For example, '<page_label>, <book_label>'
        can be an entry of the array.

    :See Also: :meth:`vsm.corpus.Corpus.view_metadata`

    **Examples**

    >>> c = random_corpus(5, 5, 1, 3, context_type='ctx', metadata=True) 
    >>> meta = c.view_metadata('ctx')
    >>> meta
    array([(2, 'ctx_0'), (3, 'ctx_1'), (4, 'ctx_2'), (5, 'ctx_3')], 
          dtype=[('idx', '<i8'), ('ctx_label', '|S5')])
    >>> def_label_fn(meta)
    array(['ctx_0', 'ctx_1', 'ctx_2', 'ctx_3'], 
          dtype='|S5')
    """
    names = [name for name in metadata.dtype.names if name.endswith('_label')]
    labels = [', '.join([x[n] for n in names]) for x in metadata]
    
    return np.array(labels)


def doc_label_name(context_type):
    """
    Takes a string `context_type` and makes that a standard 'label' string.
    This functions is useful for figuring out the specific label name in 
    metadata given the context_type.

    :param context_type: A type of tokenization.
    :type context_type: string

    :returns: label name for `context_type` as string.
    """
    return str(context_type) + '_label'


# 
# wrappers of distance functions
# 


def dist_word_word(word_or_words, corp, mat, weights=[],
                   as_strings=True, print_len=20,
                   filter_nan=True, dist_fn=angle, order='i'):
    """
    Computes and sorts the distances between a word or list of words
    and every word. If weights are provided, the word list is
    represented as the weighted average of the words in the list. If
    weights are not provided, the arithmetic mean is used.

    The columns of `mat` are assumed to be vector representations of
    words.

    """
    # Resolve `word_or_words`
    if isstr(word_or_words):
        word_or_words = [word_or_words]
    words, labels = zip(*[res_word_type(corp, w) for w in word_or_words])
    words, labels = list(words), list(labels)

    if len(words)==0:
        raise Exception('At least one target word is needed.')

    # Generate pseudo-word
    if len(weights) == 0:
        weights=None
    if issparse(mat):
        cols = mat.tocsc()[:,words].toarray()
        word = np.average(cols, weights=weights, axis=1)[np.newaxis,:]
        word = csr_matrix(word)
    else:
        cols = mat[:,words]
        word = np.average(cols, weights=weights, axis=1)[np.newaxis,:]
    
    # Compute distances
    w_arr = dist_fn(word, mat)

    if as_strings:
        w_arr = enum_sort(w_arr, indices=corp.words, field_name='word',
			filter_nan=filter_nan)
    else:
    	w_arr = enum_sort(w_arr, filter_nan=filter_nan)

    if order=='d':
        pass
    elif order=='i':
        w_arr = w_arr[::-1]
    else:
        raise Exception('Invalid order parameter.')

    # Label data
    w_arr = w_arr.view(LabeledColumn)
    w_arr.col_header = 'Words: ' + ', '.join(labels)
    w_arr.subcol_headers = ['Word', 'Distance']
    w_arr.col_len = print_len

    return w_arr


def dist_doc_doc(doc_or_docs, corp, context_type, mat, weights=[],
                 filter_nan=True, print_len=10,
                 label_fn=def_label_fn, as_strings=True,
                 dist_fn=angle, order='i'):
    """
    Computes and sorts the distances between a document or list of
    documents and every document. If weights are provided, the document
    list is represented as the weighted average of the documents in the list.
    If weights are not provided, the arithmetic mean is used.

    The columns of `mat` are assumed to represent documents.
    """
    # Resolve `doc_or_docs`
    label_name = doc_label_name(context_type)    
    if (isstr(doc_or_docs) or isint(doc_or_docs) 
        or isinstance(doc_or_docs, dict)):
        docs = [res_doc_type(corp, context_type, label_name, doc_or_docs)[0]]
    else:
        docs = [res_doc_type(corp, context_type, label_name, d)[0] 
                for d in doc_or_docs]

    if len(docs)==0:
        raise Exception('At least one target document is needed.')

    # Generate pseudo-document
    if len(weights) == 0:
        weights=None
    if issparse(mat):
        cols = mat.tocsc()[:,docs].toarray()
        doc = np.average(cols, weights=weights, axis=1)[np.newaxis,:]
        doc = csr_matrix(doc)
    else:
        cols = mat[:,docs]
        doc = np.average(cols, weights=weights, axis=1)

    # Compute distances
    d_arr = dist_fn(doc, mat)

    # Label data
    if as_strings:
        md = corp.view_metadata(context_type)
        docs = label_fn(md)
        d_arr = enum_sort(d_arr, indices=docs, field_name='doc')
    else:
        d_arr = enum_sort(d_arr, filter_nan=filter_nan)

    if order=='d':
        pass
    elif order=='i':
        d_arr = d_arr[::-1]
    else:
        raise Exception('Invalid order parameter.')

    d_arr = d_arr.view(LabeledColumn)
    # TODO: Finish this header
    d_arr.col_header = 'Documents: '
    d_arr.subcol_headers = ['Document', 'Distance']
    d_arr.col_len = print_len

    return d_arr


def dist_word_doc(word_or_words, corp, context_type, mat, weights=[], 
                  filter_nan=True, print_len=10,
                  label_fn=def_label_fn, as_strings=True,
                  dist_fn=angle, order='i'):
    """
    Computes distances between a word or a list of words to every
    document and sorts the results. The function constructs a
    pseudo-document vector from `word_or_words` and `weights`: the
    vector representation is non-zero only if the corresponding word
    appears in the list. If `weights` are not given, `1` is assigned
    to each word in `word_or_words`.
    """
    # Resolve `word_or_words`
    if isstr(word_or_words):
        word_or_words = [word_or_words]
    words, labels = zip(*[res_word_type(corp, w) for w in word_or_words])
    words, labels = list(words), list(labels)

    if len(words)==0:
        raise Exception('At least one target word is needed.')

    # Generate pseudo-document
    doc = np.zeros(mat.shape[0], dtype=np.float)
    if len(weights) == 0:
        doc[words] = np.ones(len(words))
    else:
        doc[words] = weights
    if issparse(mat):
        doc = coo_matrix(doc)

    # Compute distances
    d_arr = dist_fn(doc.T, mat)

    # Label data
    if as_strings:
        md = corp.view_metadata(context_type)
        docs = label_fn(md)
        d_arr = enum_sort(d_arr, indices=docs, field_name='doc')
    else:
        d_arr = enum_sort(d_arr, filter_nan=filter_nan)

    if order=='d':
        pass
    elif order=='i':
        d_arr = d_arr[::-1]
    else:
        raise Exception('Invalid order parameter.')

    d_arr = d_arr.view(LabeledColumn)
    # TODO: Finish this header
    d_arr.col_header = 'Words: '
    d_arr.subcol_headers = ['Document', 'Distance']
    d_arr.col_len = print_len

    return d_arr


def dist_word_top(word_or_words, corp, mat, weights=[], 
                  print_len=10, filter_nan=True, 
                  dist_fn=JS_dist, order='i'):
    """
    Computes distances between a word or a list of words to every topic
    and sorts the results. The function treats the query words as a pseudo-topic
    that assign to those words non-zero probability masses specified by `weight`.
    Otherwise equal probability is assigned to each word in `word_or_words`. 
    """
    # Resolve `word_or_words`
    if isstr(word_or_words):
        word_or_words = [word_or_words]
    words, labels = zip(*[res_word_type(corp, w) for w in word_or_words])
    words, labels = list(words), list(labels)

    if len(words)==0:
        raise Exception('At least one target word is needed.')

    # Generate pseudo-topic
    top = np.zeros(mat.shape[0], dtype=np.float)
    if len(weights) == 0:
        top[words] = np.ones(len(words))
    else:
        top[words] = weights

    # Compute distances
    k_arr = dist_fn(top, mat)
    k_arr = enum_sort(k_arr, filter_nan=filter_nan)

    if order=='d':
        pass
    elif order=='i':
        k_arr = k_arr[::-1]
    else:
        raise Exception('Invalid order parameter.')

    # Label data
    k_arr = k_arr.view(LabeledColumn)
    k_arr.col_header = 'Words: ' + ', '.join(labels)
    k_arr.subcol_headers = ['Topic', 'Distance']
    k_arr.col_len = print_len

    return k_arr


def dist_top_doc(topic_or_topics, mat, corp, context_type, weights=[],
                 print_len=10, filter_nan=True, 
                 label_fn=def_label_fn, as_strings=True,
                 dist_fn=JS_dist, order='i'):
    """
    Takes a topic or list of topics (by integer index) and returns a
    list of documents sorted by distance.
    
    The columns of `mat` are assumed to represent documents.    
    """
    topics = res_top_type(topic_or_topics)

    if len(topics)==0:
        raise Exception('At least one target topic is needed.')

    # Generate pseudo-document
    doc = np.zeros(mat.shape[0], dtype=np.float)
    if len(weights) == 0:
        doc[topics] = np.ones(len(topics))
    else:
        doc[topics] = weights
    doc /= doc.sum()

    # Compute similarites/divergences
    d_arr = dist_fn(doc, mat)

    # Label data
    if as_strings:
        md = corp.view_metadata(context_type)
        docs = label_fn(md)
        d_arr = enum_sort(d_arr, indices=docs, field_name='doc')
    else:
        d_arr = enum_sort(d_arr, filter_nan=filter_nan)

    if order=='d':
        pass
    elif order=='i':
        d_arr = d_arr[::-1]
    else:
        raise Exception('Invalid order parameter.')

    d_arr = d_arr.view(LabeledColumn)
    d_arr.col_header = 'Topics: ' + ', '.join([str(t) for t in topics])
    d_arr.subcol_headers = ['Document', 'Distance']
    d_arr.col_len = print_len

    return d_arr


def dist_top_top(mat, topic_or_topics, weights=[], 
                 print_len=10, filter_nan=True, 
                 dist_fn=JS_dist, order='i'):
    """
    Takes a topic or list of topics (by integer index) and returns
    a list of topics sorted by the distances between a given topic
    and every topic.

    The columns of `mat` are assumed to be probability distributions
    (namely, topics).
    """
    topics = res_top_type(topic_or_topics)

    if len(topics)==0:
        raise Exception('At least one target topic is needed.')

    # Generate pseudo-topic
    if len(weights) == 0:
        weights = None
    top = np.average(mat[:,topics], weights=weights, axis=1)

    # Compute distances
    k_arr = dist_fn(top, mat)
    k_arr = enum_sort(k_arr, filter_nan=filter_nan)

    if order=='d':
        pass
    elif order=='i':
        k_arr = k_arr[::-1]
    else:
        raise Exception('Invalid order parameter.')

    # Label data
    k_arr = k_arr.view(LabeledColumn)
    k_arr.col_header = 'Topics: ' + ', '.join([str(t) for t in topics])
    k_arr.subcol_headers = ['Topic', 'Distance']
    k_arr.col_len = print_len

    return k_arr


def dismat_word(word_list, corp, mat, dist_fn=angle):
    """
    Calculates a distance matrix for a given list of words.

    The columns of `mat` are assumed to represent words.
    """
    indices, words = zip(*[res_word_type(corp, word) 
                           for word in word_list])
    indices, words = np.array(indices), np.array(words)

    mat = mat[:,indices]

    dm = dist_fn(mat.T, mat)
    if np.isscalar(dm):
        dm = np.array([dm])
    dm = dm.view(IndexedSymmArray)
    dm.labels = words
    
    return dm


#TODO: Abstract the label creation correctly
def dismat_doc(doc_list, corp, context_type, mat, dist_fn=angle):
    """
    Calculates a distance matrix for a given list of documents.
    
    The columns of `mat` are assumed to be probability distributions
    (namely, over topics).
    """
    label_name = doc_label_name(context_type)

    indices, labels = zip(*[res_doc_type(corp, context_type, label_name, doc) 
                            for doc in doc_list])
    indices, labels = np.array(indices), np.array(labels)

    mat = mat[:,indices]

    dm = dist_fn(mat.T, mat)
    if np.isscalar(dm):
        dm = np.array([dm])
    dm = dm.view(IndexedSymmArray)
    dm.labels = labels
    
    return dm


def dismat_top(topics, mat, dist_fn=JS_dist):
    """
    Calculates a distance matrix for a given list of topics.
    
    The columns of `mat` are assumed to be probability distributions
    (namely, topics).
    """
    mat = mat[:,topics]

    dm = dist_fn(mat.T, mat)
    if np.isscalar(dm):
        dm = np.array([dm])
    dm = dm.view(IndexedSymmArray)
    dm.labels = [str(k) for k in topics]
    
    return dm



import numpy as np

def gen_colors(clusters):
    """
    Takes 'clusters' and creates a list of colors so a cluster has a color.
    
    :param clusters: A flat list of integers where an integer represents which
        cluster the information belongs to.
    :type clusters: list
    
    :returns: colorm : list
        A list of colors obtained from matplotlib colormap cm.hsv. The
        length of 'colorm' is the same as the number of distinct
        clusters.
    """
    import matplotlib.cm as cm
    
    n = len(set(clusters))
    colorm = [cm.hsv(i * 1.0 /n, 1) for i in xrange(n)]
    return colorm


def plot_clusters(arr, labels, clusters=[], size=[]):
    """	
    Takes 2-dimensional array(simmat), list of clusters, list of labels,
    and list of marker size. 'clusters' should be a flat list which can be
    obtained from cluster_topics(by_cluster=False).
    Plots each clusters in different colors.
    
    :type arr: 2-dimensional array
    :param arr: Array has x, y coordinates to be plotted on a 2-dimensional
        space.
    
    :param labels: List of labels to be displayed in the graph. 
    :type labels: list
    
    :param clusters: A flat list of integers where an integer represents which
        cluster the information belongs to. If not given, it returns a
        basic plot with no color variation. Default is an empty list.
    :type clusters: list, optional
    
    :param size: List of markersize for points where markersize can note the
        importance of the point. If not given, 'size' is a list of
        fixed markersize, 40. Default is an empty list.
    :type size: list, optional

    :returns: plt : maplotlit.pyplot object
        A graph with scatter plots from 'arr'.
    """
    import matplotlib.pyplot as plt

    n = arr.shape[0]
    X = arr[:,0]
    Y = arr[:,1]

    if len(size) == 0:
        size = [40 for i in xrange(n)]
        
    fig = plt.figure(figsize=(10,10))
    ax = plt.subplot(111)

    if len(clusters) == 0:
        plt.scatter(X, Y, size)
    else:	
        colors = gen_colors(clusters)
        colors = [colors[i] for i in clusters]

        for i in xrange(n):
            plt.scatter(X[i], Y[i], size, color=colors[i])

    ax.set_xlim(np.min(X) - .1, np.max(X) + .1)
    ax.set_ylim(np.min(Y) - .1, np.max(Y) + .1)
    ax.set_xticks([])
    ax.set_yticks([])

    for label, x, y in zip(labels, X, Y):
        plt.annotate(label, xy = (x, y), xytext=(-2, 3), 
                     textcoords='offset points', fontsize=10)

    plt.show()



from inpho.model import *

ideas = Session.query(Idea)
words_int = dict([(idea.label, idea.ID) for idea in ideas.all()])

def inpho_word_tokenize(document, terms=None):
    if terms is None:
        terms = ideas[:]
    occurrences = []
    
    # iterate over terms to be scanned
    for term in terms:
        # build list of search patterns starting with label
        for pattern in term.patterns:
            try:
                if re.search(pattern, document, flags=re.IGNORECASE):
                    occurrences.append(str(term.ID))
                    break
            except re.error:
                logging.warning('Term %d (%s) pattern "%s" failed' % 
                                (term.ID, term.label, pattern))

    return occurrences





"""
[General Documentation about :mod:`model` classes]
"""
from __future__ import absolute_import

from . import beaglecomposite
from .beaglecomposite import *
from . import beaglecontext
from .beaglecontext import *
from . import beagleenvironment
from .beagleenvironment import *
from . import beagleorder
from .beagleorder import *
from . import lda
from .lda import *
from . import ldacgsseq
from .ldacgsseq import *
from . import ldacgsmulti
from .ldacgsmulti import *
from . import lsa
from .lsa import *
from . import tf
from .tf import *
from . import tfidf
from .tfidf import *


__all__ = beaglecomposite.__all__[:]
__all__ += beaglecontext.__all__
__all__ += beagleenvironment.__all__
__all__ += beagleorder.__all__
__all__ += lda.__all__
__all__ += ldacgsseq.__all__
__all__ += ldacgsmulti.__all__
__all__ += lsa.__all__
__all__ += tf.__all__
__all__ += tfidf.__all__



import numpy as np

from vsm.spatial import angle
from vsm.exceptions import *

from vsm.viewer.wrappers import *


__all__ = ['BeagleViewer']


class BeagleViewer(object):
    """
    A class for viewing BEAGLE models.
    """
    
    def __init__(self, corpus, model):
        """ 
        Initialize BeagleViewer.

        :param corpus: Source of observed data.
        :type corpus: :class:`Corpus`
    
        :param model: One of the Beagle objects.
        :type model: BEAGLE model
        """
        self.corpus = corpus
        self.model = model

    def dist_word_word(self, word_or_words, weights=[], 
                       filter_nan=True, print_len=10, as_strings=True,
                       dist_fn=angle, order='i'):
        """
        Computes and sorts the distances between word(s) and every word.

        :param word_or_words: Query word(s) to which distances are calculated.
        :type word_or_words: string or list of strings
        
        :param weights: Specify weights for each query word in `word_or_words`. 
            Default uses equal weights (i.e. arithmetic mean)
        :type weights: list of floating point, optional
        
        :param filter_nan: If `True` not a number entries are filtered.
            Default is `True`.
        :type filter_nan: boolean, optional

        :param print_len: Number of words to be displayed. Default is 10.
        :type print_len: int, optional

        :param as_strings: If `True`, returns a list of words as strings rather
            than their integer representations. Default is `True`.
        :type as_strings: boolean, optional

        :param dist_fn: A distance function from functions in vsm.spatial. 
            Default is :meth:`angle`.
        :type dist_fn: string, optional

        :param order: Order of sorting. 'i' for increasing and 'd' for
            decreasing order. Default is 'i'.
        :type order: string, optional

        :returns: an instance of :class:`LabeledColumn`.
            A 2-dim array containing words and their distances to 
            `word_or_words`. 
        
        :See Also: :meth:`vsm.viewer.wrappers.dist_word_word`
        """
        return dist_word_word(word_or_words, self.corpus, 
                              self.model.matrix.T, weights=weights, 
                              filter_nan=filter_nan, 
                              print_len=print_len, as_strings=True,
                              dist_fn=dist_fn, order=order)


    @deprecated_meth("dist_word_word")
    def sim_word_word(self, word_or_words, weights=[], 
                       filter_nan=True, print_len=10, as_strings=True,
                       dist_fn=angle, order='i'):

        self.sim_word_word.__func__.__doc__ = dist_word_word.__doc__
        pass

    
    @deprecated_meth("dismat_word")
    def simmat_word(self, word_list, dist_fn=angle):
        pass

    def dismat_word(self, word_list, dist_fn=angle):        
        """
        Calculates a distance matrix for a given list of words.

        :param word_list: A list of words whose distance matrix is to be
            computed.
        :type word_list: list of strings

        :param dist_fn: A distance function from functions in vsm.spatial. 
            Default is :meth:`angle`.
        :type dist_fn: string, optional

        :returns: an instance of :class:`IndexedSymmArray`.
            A n x n matrix containing floats where n is the number of words
            in `word_list`.
        
        :See Also: :meth:`vsm.viewer.wrappers.dismat_word`
        """

        return dismat_word(word_list, self.corpus, 
                           self.model.matrix.T, dist_fn=dist_fn)



    # # This is a quick adaptation of the isomap_docs function from
    # # ldagibbsviewer. This should be abstracted and moved to
    # # similarity.py or something equivalent.
    # def isomap_words(self, words, weights=[], thres=.8,
    #                  n_neighbors=5, scale=True, trim=20):
    #     """
    #     """
    #     from sklearn import manifold
    #     from math import ceil
    #     from vsm.ext.clustering.plotting import (
    #         gen_colors as _gen_colors_,
    #         plot_clusters as _plot_clusters_)

    #     # create a list to be plotted
    #     word_list = self.dist_word_word(words, weights=weights)

    #     # cut down the list by the threshold
    #     labels, size = zip(*[(w,s) for (w,s) in word_list if s < thres])
    #     print size
    #     # calculate coordinates
    #     dismat = self.dismat_words(labels)
    #     dismat = np.clip(dismat, 0, 2)     # cut off values outside [0, 1]
    #     imap = manifold.Isomap(n_components=2, n_neighbors=n_neighbors)
    #     pos  = imap.fit(dismat).embedding_

    #     # set graphic parameters
    #     # - scale point size
    #     if scale:
    #         size = [s+0.5 if s == 0 else s for s in size] # for given word which has 0.0
    #         # value to be visible.
    #         size = [s**2*150 for s in size] 
    #     else:
    #         size = np.ones_like(size) * 50
    #     # - trim labels
    #     if trim:
    #         labels = [lab[:trim] for lab in labels]

    #     # hack for unidecode issues in matplotlib
    #     labels = [label.decode('utf-8', 'ignore') for label in labels]
        
    #     return _plot_clusters_(pos, labels, size=size)




from __future__ import division
from builtins import range

import numpy as np
from scipy.sparse import csr_matrix

from vsm.model.base import BaseModel


__all__ = [ 'TfIdf' ]


class TfIdf(BaseModel):
    """
    Transforms a term-frequency model into a term-frequency
    inverse-document-frequency model.

    A TF-IDF model is term frequency model whose rows, corresponding
    to word types, are scaled by IDF values. The idea is that a word
    type which occurs in most of the contexts (i.e., documents) does
    less to distinguish the contexts semantically than does a word
    type which occurs in few of the contexts. The document frequency
    is the number of documents in which a word occurs divided by the
    number of documents. The IDF is the log of the inverse of the
    document frequency.

    As with a term-frequency model, word types correspond to matrix
    rows and contexts correspond to matrix columns.

    The data structure is a sparse float matrix.

    :See Also: :class:`vsm.model.TfSeq`, :class:`vsm.model.base`,
        :class:`scipy.sparse.coo_matrix`

    :notes:
        A zero in the matrix might arise in two ways: (1) the word type
        occurs in every document, in which case the IDF value is 0; (2)
        the word type occurs in no document at all, in which case the IDF
        value is undefined.
    """
    def __init__(self, corpus=None, context_type=None, tf_matrix=None):
        """
        Initialize TfIdf.

        :param corpus: A Corpus object containing the training data.
        :type corpus: Corpus
    
        :param context_type: A string specifying the type of context over
            which the model trainer is applied.
        :type context_type: string 

        :param tf_matrix: A matrix containing the term-frequency data.
        :type tf_matrix: scipy.sparse matrix
        """

        self.context_type = context_type
        if corpus is not None:
            self.corpus = corpus.corpus
        else:
            self.corpus = []

        if tf_matrix is None:
            self.matrix = csr_matrix([], dtype=np.float64)
        else:
            self.matrix = tf_matrix.copy()
            self.matrix = self.matrix.tocsr()
            self.matrix = self.matrix.astype(np.float64)

        self.undefined_rows = []


    def train(self):
        """
        Computes the IDF values for the input term-frequency matrix,
        scales the rows by these values and stores the results in
        `self.matrix`.
        """
        if self.matrix.size > 0:
            n_docs = np.float64(self.matrix.shape[1])
            
            for i in range(self.matrix.indptr.shape[0] - 1):

                start = self.matrix.indptr[i]
                stop = self.matrix.indptr[i + 1]

                if start == stop:
                    self.undefined_rows.append(i)
                else:
                    row = self.matrix.data[start:stop]
                    row *= np.log(n_docs / np.count_nonzero(row))
                    start = stop

    @staticmethod
    def from_tf(tf_model):
        """
        Takes a `Tf` model object and generates a `TfIdf` model.
        """
        model = TfIdf(tf_matrix=tf_model.matrix)
        model.corpus = tf_model.corpus
        model.context_type = tf_model.context_type
        return model



"""
Functions for splitting lists and arrays
"""


from builtins import range
import numpy as np


__all__ = ['split_corpus', 'mp_split_ls', 'split_documents']



def split_corpus(arr, indices):
    """
    Splits the given array by the indices into list of sub-arrays.
    
    :param arr: An array to be split.
    :type arr: array
    :param indices: 1-dimensional array of integers that indicates 
        where the array is split.
    :type indices: array
   
    :returns: A list of sub-arrays split at the indices.
   
    **Examples**

    >>> arr = np.arange(8)
    >>> indices = np.array([2,4,7])
    >>> split_corpus(arr, indices)
    [array([0,1]), array([2,3]), array([4,5,6]), array([7])]
    """
    if len(indices) == 0:
        return arr

    if isinstance(indices, list):
        indices = np.array(indices)

    out = np.split(arr, indices)
    
    if (indices >= len(arr)).any():
        out = out[:-1]
    try:
        for i in range(len(out)):
            if out[i].size == 0:
                out[i] = np.array([], dtype=arr.dtype)
    except AttributeError:
        for i in range(len(out)):
            if out[i].size == 0:
                out[i] = np.array([])

    return out



def mp_split_ls(ls, n):
    """
    Split list into an `n`-length list of arrays.

    :param ls: List to be split.
    :type ls: list

    :param n: Number of splits.
    :type n: int

    :returns: List of arrays whose length is 'n'.

    **Examples**
    >>> ls = [1,5,6,8,2,8]
    >>> mp_split_ls(ls, 4)
    [array([1, 5]), array([6, 8]), array([2]), array([8])]
    """
    return np.array_split(ls, min(len(ls), n))


def split_documents(corpus, indices, n_partitions):
    """
    """
    docs = [(0, indices[0])]
    for i in range(len(indices)-1):
        docs.append((indices[i], indices[i+1]))
    docs = np.array(docs, dtype='i8, i8')

    corpus_chunks = np.array_split(corpus, n_partitions)
    chunk_indices = np.cumsum([len(chunk) for chunk in corpus_chunks])
    doc_indices = np.searchsorted(indices, chunk_indices, side='right')
    doc_partitions = np.split(docs, doc_indices[:-1])

    return doc_partitions



from __future__ import print_function
from __future__ import absolute_import
from future import standard_library
from functools import reduce
standard_library.install_aliases()
from builtins import zip
from builtins import str
from builtins import range
import os
import shutil
import tempfile
import multiprocessing as mp
import pickle as cpickle

import numpy as np
from numpy import dual

from vsm.spatial import rand_pt_unit_sphere
from vsm.model.base import BaseModel


__all__ = [ 'BeagleOrderSeq', 'BeagleOrderMulti' ]



def two_rand_perm(n, seed=None):
    """
    """
    np.random.seed(seed)
    
    perm1 = np.random.permutation(n)

    while True:
        perm2 = np.random.permutation(n)

        if not (perm2 == perm1).all():
            break
        
    return perm1, perm2


def mk_b_conv(n, rand_perm=None):
    """
    """
    if rand_perm is None:
        rand_perm = two_rand_perm(n)
    
    def b_conv(v1, v2):
        
        w1 = dual.fft(v1[rand_perm[0]])
        w2 = dual.fft(v2[rand_perm[1]])

        return np.real_if_close(dual.ifft(w1 * w2))

    return b_conv


def ngram_slices(i, n, l):
    """
    Given index i, n-gram width n and array length l, returns slices
    for all n-grams containing an ith element
    """
    out = []

    a = i - n + 1
    if a < 0:
        a = 0

    b = i + 1
    if b + n > l:
        b = l - n + 1

    d = b - a

    for k in range(d):
        start = a + k
        stop = start + n
        out.append(slice(start, stop))

    return out


def reduce_ngrams(fn, a, n, i, flat=True):
    """
    Given array a, reduce with fn all n-grams with width n and less and
    that contain the element at i.

    Memoizes.
    """
    m = n if n < a.shape[0] else a.shape[0]

    out = { 1: { i: a[i] } }
    
    for j in range(2, m + 1):

        slices = ngram_slices(i, j, a.shape[0])
        init = slices[0]
        out[j] = { init.start: reduce(fn, a[init]) }

        for s in slices[1:]:
            prev = out[j - 1][s.start]
            out[j][s.start] = fn(prev, a[s.stop - 1])

    # Delete the 1-gram
    del out[1]

    if flat:
        out = [v for d in out.values() for v in d.values()]
    
    return out


class BeagleOrderSeq(BaseModel):
    """
    `BeagleOrderSeq` stores word order information in the context.
    """

    def __init__(self, corpus, env_matrix, context_type='sentence',
                 psi=None, rand_perm=None, lmda =7):
        """ 
        Initialize BeagleOrderSeq.
        
        :param corpus: Soure of observed data.
        :type corpus: Corpus

        :param env_matrix: BEAGLE environment matrix.
        :type env_matrix: 2-D array

        :param context_type: Name of tokenization stored in `corpus` whose
            tokens will be treated as documents. Default is `sentence`.
        :type context_type: string, optional

        :param psi:  
        :type psi: int, optional

        :param rand_perm:  
        :type rand_perm: boolean, optional

        :param lmda:  
        :type lmda: int, optional
        """
        self.context_type = context_type
        self.sents = corpus.view_contexts(context_type)
        self.env_matrix = env_matrix

        self.b_conv = mk_b_conv(env_matrix.shape[1], rand_perm)
        
        if psi is None:
            self.psi = rand_pt_unit_sphere(env_matrix.shape[1])
        else:
            self.psi = psi

        self.lmda = lmda


    def train(self):
        """
        Trains the model.
        """
        self.matrix = np.zeros_like(self.env_matrix)

        for sent in self.sents:
            env = self.env_matrix[sent]
            
            for i in range(sent.size):
                sent_vecs = env.copy() 
                sent_vecs[i, :] = self.psi[:]

                conv_ngrams = reduce_ngrams(self.b_conv, sent_vecs, self.lmda, i)
                    
                ord_vec = np.sum(conv_ngrams, axis=0)
                self.matrix[sent[i], :] += ord_vec



class BeagleOrderMulti(BaseModel):
    """
    `BeagleOrderSeq` stores word order information in the context.
    """
    def __init__(self, corpus, env_matrix, context_type='sentence',
                 psi=None, rand_perm=None, lmda =7):
        """    
        Initialize BeagleOrderMulti.

        :param corpus: Source of observed data.
        :type corpus: Corpus

        :param env_matrix: BEAGLE environement matrix.
        :type env_matrix: 2-D array

        :param context_type: Name of tokenization stored in `corpus` whose
            tokens will be treated as documents. Default is `sentence`.
        :type context_type: string, optional

        :param psi:  
        :type psi: int, optional

        :param rand_perm:  
        :type rand_perm: boolean, optional

        :param lmda:  
        :type lmda: int, optional
        """
        self.context_type = context_type
        self.sents = corpus.view_contexts(context_type)
        self.dtype = env_matrix.dtype

        global _shape 
        _shape = mp.Array('i', 2, lock=False)
        _shape[:] = env_matrix.shape

        print('Copying env matrix to shared mp array')
        global _env_matrix
        _env_matrix = mp.Array('d', env_matrix.size, lock=False)
        _env_matrix[:] = env_matrix.ravel()[:]

        #TODO: Convert this to a shared data structure
        global _b_conv 
        _b_conv = mk_b_conv(env_matrix.shape[1], rand_perm)

        global _psi        
        _psi = mp.Array('d', _shape[1], lock=False)
        if psi is None:
            _psi[:] = rand_pt_unit_sphere(env_matrix.shape[1])[:]
        else:
            _psi[:] = psi[:]

        global _lmda
        _lmda = mp.Value('i', lock=False)
        _lmda =lmda


    def train(self, n_procs=2):
        """
        Trains the model using `n_procs` processors.

        :param n_procs: Number of processors. Default is 2.
        :type n_procs: int, optional
        """
        sent_lists = np.array_split(self.sents, n_procs-1)
        if len(sent_lists) != n_procs:
            sent_lists = np.array_split(self.sents, n_procs)

        tmp_dir = tempfile.mkdtemp()
        tmp_files = [os.path.join(tmp_dir, 'tmp_' + str(i))
                     for i in range(len(sent_lists))]

        sent_lists = list(zip(sent_lists, tmp_files))
        del self.sents


        try:
            print('Forking')
            # For debugging
            # tmp_files = map(mpfn, sent_lists)
            
            p = mp.Pool(n_procs)
            tmp_files = p.map(mpfn, sent_lists, 1)
            p.close()

            print('Reducing')
            self.matrix = np.zeros(tuple(_shape), dtype=self.dtype)

            for filename in tmp_files:

                with open(filename, 'rb') as f:
                    result = cpickle.load(f)

                for k,v in result.items():
                    self.matrix[k, :] += v

        finally:
            print('Removing {}'.format(tmp_dir))
            shutil.rmtree(tmp_dir)
        

def mpfn(sents_filename):
    """
    """
    (sents, filename) = sents_filename
    result = dict()

    for sent in sents:

        env = np.empty((sent.size, _shape[1]), dtype=np.float64)
        for i,w in enumerate(sent):
            env[i, :] = _env_matrix[w*_shape[1]: (w+1)*_shape[1]]

        for i,w in enumerate(sent):
            sent_vecs = env.copy()
            sent_vecs[i, :] = _psi[:]

            conv_ngrams = reduce_ngrams(_b_conv, sent_vecs, _lmda, i)

            ord_vec = np.sum(conv_ngrams, axis=0)

            if w in result:
                result[w] += ord_vec
            else:
                result[w] = ord_vec
            
    with open(filename, 'wb') as f:
        cpickle.dump(result, f)

    return filename



from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import zip
from builtins import range
import os
import shutil
import re
import logging

import numpy as np

from unidecode import unidecode

from nltk.corpus import wordnet as wn
import enchant

from vsm.extensions.corpusbuilders.util import filter_by_suffix


import json
from time import sleep
from urllib.request import urlopen
from urllib.parse import quote_plus

def metadata(id, sleep_time=1):
    """
    Given a HTRC ID, download the volume metadata from the Solr index.

    
    :param id: HTRC volume id.
    :type id: string

    :param sleep_time: Sleep time to prevent denial of service
    :type sleep_time: int in seconds, default: 1

    :returns: dict

    """
    solr ="http://chinkapin.pti.indiana.edu:9994/solr/meta/select/?q=id:%s" % id
    solr += "&wt=json" ## retrieve JSON results
    # TODO: exception handling
    if sleep_time:
        sleep(sleep_time) ## JUST TO MAKE SURE WE ARE THROTTLED
    try:
        data = json.load(urlopen(solr))
        print(id)
        return data['response']['docs'][0]
    except ValueError as IndexError:
        print("No result found for " + id) 
        return dict()

def proc_htrc_coll(coll_dir, ignore=['.json', '.log', '.err']):
    """
    Given a collection, cleans up plain page files for books in the collection.

    :param coll_dir: The path for collection.
    :type coll_dir: string
    
    :param ignore: List of file extensions to ignore in the directory.
    :type ignore: list of strings, optional

    :returns: None

    :See Also: :meth: proc_htrc_book
    """
    books = os.listdir(coll_dir)
    books = filter_by_suffix(books, ignore)
    
    for book in books:
        # For debugging
        # if book == 'uc2.ark+=13960=t1zc80k1p':
        # if book == 'uc2.ark+=13960=t8tb11c8g':
        # if book == 'uc2.ark+=13960=t74t6gz6r':
        proc_htrc_book(book, coll_dir, ignore=ignore)
            


def proc_htrc_book(book, coll_dir, ignore=['.json', '.log', '.err']):
    """
    Cleans up page headers, line breaks, and hyphens for all plain pages in the book directory. 
    Creates a log file for debugging purposes.  
  
    :param book: The name of the book directory in coll_dir.
    :type book: string
    
    :param coll_dir: The path for collection.
    :type coll_dir: string
    
    :param ignore: List of file extensions to ignore in the directory.
    :type ignore: list of strings, optional

    :returns: None

    :See Also: :meth: rm_pg_headers, :meth: rm_lb_hyphens
    """
    book_root = os.path.join(coll_dir, book)

    logger = logging.getLogger(book)
    logger.setLevel(logging.INFO)
    log_file = os.path.join(coll_dir, book + '-raw-proc.log')
    handler = logging.FileHandler(log_file, mode='w')
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    print('Processing', book_root)

    try:
        rm_pg_headers(book_root, logger, ignore=ignore)
        rm_lb_hyphens(book_root, logger, ignore=ignore)
    finally:
        handler.close()



def rm_lb_hyphens(plain_root, logger, ignore=['.json', '.log', '.err']):
    """
    Looks for a hyphen followed by whitespace or a line break.

    Reconstructs word and checks to see if the result exists in either
    WordNet or the OS's default spellchecker dictionary. If so,
    replaces fragments with reconstructed word.
    
    :param plain_root: The name of the directory containing plain-text 
        files.
    :type plain_root: string
    
    :param logger: Logger that handles logging for the given directory.
    :type logger: Logger
    
    :param ignore: List of file extensions to ignore in the directory.
    :type ignore: list of strings, optional

    :returns: None
    """

    try:
        d = enchant.Dict('en_US')
    except ImportError:
        d = None

    def recon(match_obj):
        rc_word = match_obj.group(1) + match_obj.group(2)
        
        if wn.synsets(rc_word) or (d and d.check(rc_word)):
            logger.info('\nbook: %s\nreconstructed word:\n%s\n',
                         plain_root, rc_word)
            return rc_word
        
        logger.info('\nbook: %s\nignored expression:\nleft: %s\nright: %s\n',
                     plain_root, match_obj.group(1), match_obj.group(2))

        return match_obj.group(0)

    def inner(s):
        lb_hyphenated = re.compile(r'(\w+)-\s+(\w+)')
        return lb_hyphenated.sub(recon, s)
    
    page_files = os.listdir(plain_root)
    page_files = filter_by_suffix(page_files, ignore)

    for i, page_file in enumerate(page_files):
        filename = os.path.join(plain_root, page_file)

        with open(filename, 'w+') as f:
            page = f.read()
            page = inner(page)
            f.seek(0)
            f.write(page)
            f.truncate()



def rm_pg_headers(plain_root, logger, bound=1, ignore=['.json', '.log', '.err']):
    """
    Tries to detect repeated page headers (e.g., chapter titles). If
    found, removes them.

    The routine takes the first non-empty lines of text, strips them
    of numbers and punctuation and computes frequencies. If frequency
    for the reduced string exceeds `bound`, the corresponding first
    lines are considered headers.
    
    :param plain_root: The name of the directory containing plain-text 
        files.
    :type plain_root: string
    
    :param logger: Logger that handles logging for the given directory.
    :type logger: Logger
    
    :param bound: Number of frequency of a reduced string. If the string
        appears more than `bound`, then the corresponding first lines are
        considered headers. Default is 1.
    :param bound: int, optional

    :param ignore: List of file extensions to ignore in the directory.
    :type ignore: list of strings, optional

    :returns: None
    """
    page_files = os.listdir(plain_root)
    page_files = filter_by_suffix(page_files, ignore)

    # Get first non-empty lines
    first_lines = []
    fl = re.compile(r'^\s*([^\n]*)\n')
    
    for page_file in page_files:
        page_file = os.path.join(plain_root, page_file)

        with open(page_file, 'r') as f:
            page = f.read()

        first_line = fl.match(page)
        if first_line == None:
            first_lines.append('')
        else:
            first_lines.append(first_line.group(0))

    # Remove capitalization, roman numerals for numbers under 50,
    # punctuation, arabic numerals from first lines
    for i in range(len(first_lines)):
        line = first_lines[i]
        line = line.lower()

        # An overzealous arabic numeral detector (OCR errors include
        # `i` for `1` for example)
        line = re.sub(r'\b\S*\d+\S*\b', '', line)

        # Roman numerals i to xxxix
        line = re.sub(r'\b(x{0,3})(ix|iv|v?i{0,3})\b', '', line)

        # Collapse line to letters only
        line = re.sub(r'[^a-z]', '', line)
        first_lines[i] = (first_lines[i], line)

    freqs = dict()
    for line, reduced in first_lines:
        if reduced in freqs:
            freqs[reduced] += 1
        else:
            freqs[reduced] = 1
    
    for i, page_file in enumerate(page_files):
        filename = os.path.join(plain_root, page_file)
        line, reduced = first_lines[i]

        if freqs[reduced] > bound:
            with open(filename, 'r') as f:
                page = f.read()
            if page:
                logger.info(u'\nbook: %s\nfile: %s\nremoved header:\n%s\n',
                             unidecode(plain_root), unidecode(page_file), unidecode(line))
            page = fl.sub('', page)

            with open(filename, 'w') as f:
                f.write(page)


def htrc_load_metadata_1315():
    """
    Loads hathitrust metadata for the 1315 volumes.
    """
    import os
    import json

    filename = ('/var/inphosemantics/data/20130101/htrc-anthropomorphism-1315/'
                'htrc-1315-metadata.json')

    with open(filename) as f:
        metadata = json.load(f)

    return metadata


def htrc_load_metadata_86():
    """
    Loads hathitrust metadata for the 86 volumes.
    """
    import os
    import json

    filename = ('/var/inphosemantics/data/20130101/htrc-anthropomorphism-86/'
                'htrc-anthropomorphism-86-metadata.json')

    with open(filename) as f:
        metadata = json.load(f)

    return metadata


def htrc_get_titles(metadata, vol_id):
    """
    Gets titles of the volume given the metadata from a json file
    and volume id.
    """
    try:
        md = metadata[vol_id]
        return md[list(md.keys())[0]]['titles']

    except KeyError:
        print('Volume ID not found:', vol_id)
        raise


def htrc_label_fn_86(metadata):
    """
    A customized label function for hathitrust 86 volumes.
    It loads the metadata of htrc 86 and returns labels that consist of
    file names which are pages and book titles.

    :param metadata: Strucutred array that has 'file' and 'book_label' field.
        Most likely the output of Corpus.view_metadata('page').
    :type metadata: array

    :returns: An array of labels that consist of file names and book titles.

    :See Also: :meth: Corpus.view_metadata
    """
    md = htrc_load_metadata_86()

    files = metadata['file']
    titles = []
    for v in metadata['book_label']:
        title = unidecode(htrc_get_titles(md, v)[0])
        if len(title) > 15:
            title = title[:15]
        titles.append(title)
    
    labels = ['{0}, {1}'.format(t,f) for (t,f) in zip(titles, files)]
    
    return np.array(labels)



def htrc_label_fn_1315(metadata):
    """
    A customized label function for hathitrust 1315 volumes.
    It loads the metadata of htrc 86 and returns book title labels.

    :param metadata: Strucutred array that has 'book_label' field.
        Most likely the output of Corpus.view_metadata('book').
    :type metadata: array

    :returns: An array of labels that consist of book titles.

    :See Also: :meth: Corpus.view_metadata
    """
    md = htrc_load_metadata_1315()

    titles = []
    for v in metadata['book_label']:
        title = unidecode(htrc_get_titles(md, v)[0])
        if len(title) > 60:
            title = title[:60]
        titles.append(title)
    
    return np.array(titles)


def htrc_find_duplicates(metadata, vol_list):
    """
    Takes metadata and a list of volumes and finds duplicates
    amongst the volumes.

    :param metadata: Dictionary of metadata. Output of
        htrc_load_metadata_86() or htrc_load_metadata_1315()
        is acceptable.
    :type metadata: dictionary

    :param vol_list: List of volumes. An example of a volume is
        'uc2.ark+=13960=t73t9h556'.
    :type vol_list: list

    :returns: indices : the indices of duplicates from `vol_list`.
    
    :See Also: :meth: htrc_load_metadata_86, :meth: htrc_load_metadata_1315
    """
    record_ids = [list(metadata[vol].keys())[0] for vol in vol_list]
    
    mem, indices = [], []

    for i,r in enumerate(record_ids):
        if r in mem:
            indices.append(i)
        elif r in record_ids[:i]:
            indices.append(i)
            mem.append(r)

    return indices


def add_link(s):
    """
    if `s` is a url, then adds anchor tags for html representation in ipynb.
    """
    if s.startswith('http'):
        a = '<a href="{0}" target="_blank">'.format(s)
        a += s
        a += '</a>'
        return a

def htrc_label_link_fn_86(metadata):
    """
    A customized label function for hathitrust 86 volumes.
    It loads the metadata of htrc 86 and returns pages, book title labels, and page urls
    that open up the page on hathitrust website.

    :param metadata: Strucutred array that has 'file', and 'book_label' field.
        Most likely the output of Corpus.view_metadata('page').
    :type metadata: array

    :returns: An array of labels for pages, books, and page urls.

    :See Also: :meth: Corpus.view_metadata, :meth: htrc_label_fn_86
    """
    md = htrc_load_metadata_86()

    files = metadata['file'] 
    titles = []
    for v in metadata['book_label']:
        title = unidecode(htrc_get_titles(md, v)[0])
        if len(title) > 1:
            title = title[:15]
        titles.append(title)
    
    names = [name for name in metadata.dtype.names if name.endswith('_label')]
    links = [add_link(x[n]) for n in names if n.endswith('_url_label') for x in metadata]

    labels = ['{0}, {1}, {2}'.format(t,f,l) for (t,f,l) in zip(titles, files, links)]
    return np.array(labels)


def htrc_label_link_fn_1315(metadata):
    """
    A customized label function for hathitrust 1315 volumes.
    It loads the metadata of htrc 1315 and returns book title labels and their urls
    that open up the book on hathitrust website.

    :param metadata: Strucutred array that has 'book_label' field.
        Most likely the output of Corpus.view_metadata('page').
    :type metadata: array

    :returns: An array of labels for books and book urls.

    :See Also: :meth: Corpus.view_metadata, :meth: htrc_label_fn_1315
    """
    md = htrc_load_metadata_1315()

    titles = []
    for v in metadata['book_label']:
        title = unidecode(htrc_get_titles(md, v)[0])
        if len(title) > 60:
            title = title[:60]
        titles.append(title)
    
    names = [name for name in metadata.dtype.names if name.endswith('_label')]
    links = [add_link(x[n]) for n in names if n.endswith('_url_label') for x in metadata]

    labels = ['{0}, {1}'.format(t,l) for (t,l) in zip(titles, links)]
    return np.array(labels)



def url_metadata(corpus, ctx_type, coll_dir):
    """
    Returns a list of urls whose order matches with the existing metadata.
    It creates url metadata that can be added to a Corpus object with
    add_metadata function in vsm.ext.corpusbuilders.util.

    :param corpus: Corpus to add url metadata to. Urls match with the existing
        metadata of `corpus`.
    :type corpus: Corpus

    :param ctx_type: A type of tokenization.
    :type ctx_type: string

    :param coll_dir: Path for the collection directory. Either htrc 86 plain
        or htrc 1315 plain directory.
    :type coll_dir: string

    :returns: md : List of urls to be added to corpus

    :See Also: :meth: add_metadata
    """

    import json
    from vsm.viewer import doc_label_name
    import re

    urls = []
    corp_md = corpus.view_metadata('book')
    book_labels = corp_md[doc_label_name('book')]
    
    for book_label in book_labels:
        coll_path = os.path.join(coll_dir, book_label)
        booklist = os.listdir(coll_path)
        book = filter_by_suffix(booklist, ignore=['.txt', '.pickle'])
        
        book_path = os.path.join(coll_path, book[0])
        with open(book_path, 'r') as f:
            d = json.load(f)
            url = ''
            li = sorted(d['items'], key=lambda k: int(k['lastUpdate']))
            url = li[-1]['itemURL']

            if ctx_type == 'book':
                urls.append( unidecode(url))
            else: # urls for pages
                page_md = corpus.view_metadata('page')
                files = [a for a in page_md['file'] if a.startswith(book_label)]
                nums = [re.findall('[1-9][0-9]*', a)[-1] for a in files]
                for i in nums: 
                    s = url + '?urlappend=%3Bseq={0}'.format(i)
                    urls.append( unidecode(s))
    return urls


def page_url(corpus, ctx_type, book_path, book_id, jsonfile):
    """
    Modified htrc_*_label_fn. The individual volumes don't have 'book' as a context type.
    """
    import json
    from vsm.viewer import doc_label_name
    import re

    urls = []
    corp_md = corpus.view_metadata('page')

    jsonpath = os.path.join(book_path, jsonfile)
    with open(jsonpath, 'r') as f:
        md = json.load(f)
        url = ''
        li = sorted(md['items'], key=lambda k: int(k['lastUpdate']))
        url = li[-1]['itemURL']
            
        if ctx_type == 'book':
            urls.append( unidecode(url))
        else: # urls for pages
            page_md = corpus.view_metadata('page')
            files = page_md[doc_label_name('page')] 
            
            nums = [re.findall('[1-9][0-9]*', a)[-1] for a in files]
            for i in nums:
                s = url + '?urlappend=%3Bseq={0}'.format(i)
                urls.append( unidecode(s))
    return urls


def add_link_(s, addee):
    """
    Returns <a href="s">addee</a> 
    For example, <a href="page_url">page_label</a>
    """
    if s.startswith('http'):
        a = '<a href="{0}" target="_blank">'.format(s)
        a += addee
        a += '</a>'
        return a

def vol_link_fn(md): 
    """
    Not a generalized function for individual htrc volume with page links
    and sentence labels. 'sentences_label' is CorpusSent.sentences which
    is an array or original sentences.

    :ref: vsm.extensions.ldasentences CorpusSent
    """
    # md == corpus.view_metadata('sentence')
    links = [add_link_(x['page_urls'], x['page_label']) for x in md]
    labels = ['{0}, {1}, {2}'.format(l, i, s) for (l,i,s) in
             zip(links, md['sentence_label'], md['sentences_label'])]

    return np.array(labels)



from builtins import str
from builtins import chr
from builtins import range
from builtins import object
from past.builtins import basestring
import re
import string

from chardet.universaldetector import UniversalDetector
import numpy as np

__all__ = ['strip_punc', 'rem_num', 'rehyph',
           'apply_stoplist', 'filter_by_suffix', 'word_tokenize',
           'sentence_tokenize', 'paragraph_tokenize', 'detect_encoding',
           'sentence_span_tokenize', 'in_place_stoplist']


def _unicode_range(start, stop):
    return u''.join([chr(i) for i in range(int(start,16), int(stop,16))])

PUNCTUATION_START = re.compile(r'^(['+string.punctuation+r'\u2000-\u206F\u3000-\u303F\uFF00-\uFFFF]*)')
PUNCTUATION_END = re.compile(r'(['+string.punctuation+r'\u2000-\u206F\u3000-\u303F\uFF00-\uFFFF]*)$')
PUNC = str(string.punctuation) + _unicode_range('2000','206F') + _unicode_range('3000', '303F') + _unicode_range('FF00', 'FFFF')
PUNC_TABLE = {ord(c): None for c in PUNC}

def strip_punc(tsent):
    """
    """
    out = []
    for word in tsent:
        w = strip_punc_word(word)
        if w:
            out.append(w)

    return out

def strip_punc_word(word):
    if isinstance(word, str):
        return word.translate(PUNC_TABLE)
    elif isinstance(word, basestring):
        return word.translate(None, PUNC.encode('utf-8'))
        

NUMS = string.digits
NUMS_TABLE =  {ord(c): None for c in NUMS}

def rem_num(tsent):
    """
    """
    #p = re.compile(r'(^\D+$)|(^\D*[1-2]\d\D*$|^\D*\d\D*$)')
    out = []
    for word in tsent:
        w = rem_num_word(word)
        if w:
            out.append(w)

    return out

def rem_num_word(word):
    if isinstance(word, str):
        return word.translate(NUMS_TABLE)
    elif isinstance(word, basestring):
        return word.translate(None, NUMS.encode('utf-8'))

def rehyph(sent):
    """
    """
    return re.sub(r'(?P<x1>.)--(?P<x2>.)', '\g<x1> - \g<x2>', sent)


BIG_TABLE = NUMS_TABLE.copy()
BIG_TABLE.update(PUNC_TABLE)
BIG_LIST = string.digits + string.punctuation
def process_word(word):
    if isinstance(word, str):
        return word.translate(BIG_TABLE)
    elif isinstance(word, basestring):
        return word.translate(None, BIG_LIST.encode('utf-8'))

def apply_stoplist(corp, nltk_stop=True, add_stop=None, freq=0, in_place=True):
    """
    Returns a Corpus object with stop words eliminated.

    :param corp: Corpus object to apply stoplist to.
    :type corp: Corpus

    :param nltk_stop: If `True` English stopwords from nltk are included
        in the stoplist. Default is `True`.
    :type nltk_stop: boolean, optional
    
    :param add_stop: list of words to eliminate from `corp` words.
        Default is `None`.
    :type add_stop: List, optional

    :param freq: Eliminates words that appear <= `freq` times. Default is
        0.
    :type freq: int

    :returns: Corpus with words in the stoplist removed.

    :See Also: :class:`vsm.corpus.Corpus`, :meth:`vsm.corpus.Corpus.apply_stoplist`
    """
    stoplist = set()
    if nltk_stop:
        import nltk
        for w in nltk.corpus.stopwords.words('english'):
            stoplist.add(w)
    if add_stop:
        for w in add_stop:
            stoplist.add(w)

    if not in_place:
        return corp.apply_stoplist(stoplist=stoplist, freq=freq)
    else:
        corp.in_place_stoplist(stoplist=stoplist, freq=freq)
        return corp

def in_place_stoplist(corp, nltk_stop=True, add_stop=None, freq=0):
    """
    Returns a Corpus object with stop words eliminated.

    :param corp: Corpus object to apply stoplist to.
    :type corp: Corpus

    :param nltk_stop: If `True` English stopwords from nltk are included
        in the stoplist. Default is `True`.
    :type nltk_stop: boolean, optional
    
    :param add_stop: list of words to eliminate from `corp` words.
        Default is `None`.
    :type add_stop: List, optional

    :param freq: Eliminates words that appear <= `freq` times. Default is
        0.
    :type freq: int

    :returns: Corpus with words in the stoplist removed.

    :See Also: :class:`vsm.corpus.Corpus`, :meth:`vsm.corpus.Corpus.apply_stoplist`
    """
    stoplist = set()
    if nltk_stop:
        import nltk
        for w in nltk.corpus.stopwords.words('english'):
            stoplist.add(w)
    if add_stop:
        for w in add_stop:
            stoplist.add(w)

    corp.in_place_stoplist(stoplist=stoplist, freq=freq)


def filter_by_suffix(l, ignore, filter_dotfiles=True):
    """
    Returns elements in `l` that does not end with elements in `ignore`,
    and filters dotfiles (files that start with '.').

    :param l: List of strings to filter.
    :type l: list

    :param ignore: List of suffix to be ignored or filtered out.
    :type ignore: list

    :param filter_dotfiles: Filter dotfiles.
    :type filter_dotfiles: boolean, default True

    :returns: List of elements in `l` whose suffix is not in `ignore`.

    **Examples**

    >>> l = ['a.txt', 'b.json', 'c.txt']
    >>> ignore = ['.txt']
    >>> filter_by_suffix(l, ignore)
    ['b.json']
    """
    filter_list = [e for e in l if not sum([e.endswith(s) for s in ignore])]
    if filter_dotfiles:
        filter_list = [e for e in filter_list if not e.startswith('.')]
    return filter_list

class _tokenizer(object):
    def tokenize(self, text):
        return text.split()

word_tokenizer = _tokenizer()
def word_tokenize(text):
    """Takes a string and returns a list of strings. Intended use: the
    input string is English text and the output consists of the
    lower-case words in this text with numbers and punctuation, except
    for hyphens, removed.

    The core work is done by NLTK's Treebank Word Tokenizer.
    
    :param text: Text to be tokeized.
    :type text: string

    :returns: tokens : list of strings
    """
    global word_tokenizer
    if word_tokenizer is None:
        import nltk
        word_tokenizer = nltk.TreebankWordTokenizer()

    text = rehyph(text)
    text = process_word(text)
    text = text.replace(u'\x00','')
    text = text.lower()
    tokens = word_tokenizer.tokenize(text)

    #process_word = lambda x: strip_punc_word(rem_num_word(word)).lower().replace(u'\x00','')
    #tokens = [process_word(word) for word in text]

    return tokens


sent_tokenizer = None
def sentence_tokenize(text):
    """
    Takes a string and returns a list of strings. Intended use: the
    input string is English text and the output consists of the
    sentences in this text.

    This is a wrapper for NLTK's pre-trained Punkt Tokenizer.
     
    :param text: Text to be tokeized.
    :type text: string

    :returns: tokens : list of strings
    """
    global sent_tokenizer
    if sent_tokenizer is None:
        import nltk
        sent_tokenizer = nltk.data.load('tokenizers/punkt/english.pickle')

    return sent_tokenizer.tokenize(text)

def sentence_span_tokenize(text):
    """
    Takes a string and returns a list of strings. Intended use: the
    input string is English text and the output consists of the
    sentences in this text.

    This is a wrapper for NLTK's pre-trained Punkt Tokenizer.
     
    :param text: Text to be tokeized.
    :type text: string

    :returns: token_spans : iterator of (start,stop) tuples.
    """
    global sent_tokenizer
    if sent_tokenizer is None:
        import nltk
        sent_tokenizer = nltk.data.load('tokenizers/punkt/english.pickle')

    return sent_tokenizer.span_tokenize(text)



def paragraph_tokenize(text):
    """
    Takes a string and returns a list of strings. Intended use: the
    input string is English text and the output consists of the
    paragraphs in this text. It's expected that the text marks
    paragraphs with two consecutive line breaks.
     
    :param text: Text to be tokeized.
    :type text: string

    :returns: tokens : list of strings
    """

    par_break = re.compile(r'[\r\n]{2,}')
    span_start = 0
    for match in par_break.finditer(text):
        span_end = match.span()[0]
        yield text[span_start:span_end]
        span_start = match.span()[1]
    yield text[span_start:]
    #return par_break.split(text)


def detect_encoding(filename):
    """
    Takes a filename and attempts to detect the character encoding of the file
    using `chardet`.
     
    :param filename: Name of the file to process
    :type filename: string

    :returns: encoding : string
    """
    detector = UniversalDetector()
    with open(filename, 'rb') as unknown_file:
        for line in unknown_file:
            detector.feed(line)
            if detector.done:
                break
    detector.close()

    return detector.result['encoding']
 



from __future__ import print_function
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import zip
from builtins import str
from builtins import range
import os
import shutil
import tempfile
import multiprocessing as mp
import pickle as cpickle

import numpy as np

from vsm.model.base import BaseModel


__all__ = [ 'BeagleContextSeq', 'BeagleContextMulti' ]


def realign_env_mat(corpus, env_corpus, env_matrix):
    """
    """
    words = corpus.words
    indices = [env_corpus.words_int[w] for w in words]
    return env_matrix[indices]



class BeagleContextSeq(BaseModel):
    """

    """
    def __init__(self, corpus, env_corpus, env_matrix, 
                context_type='sentence'):
        """
        Initialize BeagleContextSeq.

        :param corpus: Source of observed data.
        :type corpus: class:`Corpus`

        :param env_corpus: BEAGLE environment corpus.
        :type env_corpus: class:`Corpus`

        :param env_matrix: Matrix from BEAGLE environment model.
        :type env_matrix: 2-D array

        :param context_type: Name of tokenization stored in `corpus` whose
            tokens will be treated as documents. Default is `sentence`.
        :type context_type: string, optional
        """
        self.context_type = context_type
        self.sents = corpus.view_contexts(context_type)
        self.env_matrix = realign_env_mat(corpus, env_corpus, env_matrix)


    def train(self):
        """
        Trains the model.
        """
        self.matrix = np.zeros_like(self.env_matrix)

        for sent in self.sents:

            if sent.shape[0] > 1:

                left_sums = np.cumsum(self.env_matrix[sent[:-1]], axis=0)
                right_sums = np.cumsum(self.env_matrix[sent[:0:-1]], axis=0)

                for i,word in enumerate(sent):

                    if i == 0:
                        ctx_vector = right_sums[-1]

                    elif i == sent.shape[0] - 1:
                        ctx_vector = left_sums[-1]
                    
                    else:
                        ctx_vector = left_sums[i - 1] + right_sums[-i - 1]

                    self.matrix[word, :] += ctx_vector



class BeagleContextMulti(BaseModel):
    """

    """

    def __init__(self, corpus, env_corpus, env_matrix, 
                 context_type='sentence'):
        """
        Initialize BeagleContextMulti.
        
        :param corpus: Souce of observed data.
        :type corpus: class:`Corpus`

        :param env_corpus: BEAGLE environment corpus. 
        :type env_corpus: class:`Corpus`

        :param env_matrix: Matrix from BEAGLE environment model.
        :type env_matrix: 2-D array

        :param context_type: Name of tokenization stored in `corpus` whose
            tokens will be treated as documents. Default is `sentence`.
        :type context_type: string, optional
        """
        self.context_type = context_type
        self.sents = corpus.view_contexts(context_type)
        self.dtype = env_matrix.dtype
        env_matrix = realign_env_mat(corpus, env_corpus, env_matrix)

        global _shape 
        _shape = mp.Array('i', 2, lock=False)
        _shape[:] = env_matrix.shape
        
        print('Copying env matrix to shared mp array')
        global _env_matrix
        _env_matrix = mp.Array('d', env_matrix.size, lock=False)
        _env_matrix[:] = env_matrix.ravel()[:]


    def train(self, n_procs=2):
        """
        Takes an optional argument `n_procs`, number of processors,
        and trains the model on the number of processors. `n_procs`
        is 2 by default.
        
        :param n_procs: Number of processors. Default is 2.
        :type n_procs: int, optional

        :returs: `None`
        """
        sent_lists = np.array_split(self.sents, n_procs-1)
        if len(sent_lists) != n_procs:
            sent_lists = np.array_split(self.sents, n_procs)

        tmp_dir = tempfile.mkdtemp()
        tmp_files = [os.path.join(tmp_dir, 'tmp_' + str(i))
                     for i in range(len(sent_lists))]

        sent_lists = list(zip(sent_lists, tmp_files))
        del self.sents

        try:
            print('Forking')
            # For debugging
            # tmp_files = map(mpfn, sent_lists)
            
            p = mp.Pool(n_procs)
            tmp_files = p.map(mpfn, sent_lists, 1)
            p.close()

            print('Reducing')
            self.matrix = np.zeros(tuple(_shape), dtype=self.dtype)

            for filename in tmp_files:

                with open(filename, 'rb') as f:
                    result = cpickle.load(f)

                for k,v in result.items():
                    self.matrix[k, :] += v

        finally:
            print('Removing {}'.format(tmp_dir))
            shutil.rmtree(tmp_dir)



def mpfn(sents_filename):
    """
    """
    sents, filename = sents_filename
    result = dict()

    for sent in sents:
        if sent.shape[0] > 1:

            env = np.empty((sent.size, _shape[1]), dtype=np.float64)
            for i,w in enumerate(sent):
                env[i, :] = _env_matrix[w*_shape[1]: (w+1)*_shape[1]]
        
            left_sums = np.cumsum(env[:-1], axis=0)
            right_sums = np.cumsum(env[:0:-1], axis=0)

            for i,t in enumerate(sent):
                
                if i == 0:
                    ctx_vector = right_sums[-1]

                elif i == sent.shape[0] - 1:
                    ctx_vector = left_sums[-1]
                    
                else:
                    ctx_vector = left_sums[i - 1] + right_sums[-i - 1]
                   
                if t in result:
                    result[t] += ctx_vector
                else:
                    result[t] = ctx_vector

    with open(filename, 'wb') as f:
        cpickle.dump(result, f)

    return filename

